/*
Copyright (C) 2001 Lkb

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include <d3d.h>

// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
#ifdef _MSC_VER
#include <crtdbg.h>
#endif
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support

#include "ultra_mbi.h"

#include "SurfaceHandler.h"
#include "D3DRender.h"
#include "DBGConsole.h"
#include "RDP_GFX.h"

#include <ios>
#include <map>
#include <vector>
#include <functional>
#include <algorithm>
#include <ostream>
#include <fstream>
#include <strstream>

using namespace std;

extern D3DDEVICEDESC7 g_D3DDeviceDesc;
extern D3DRender* g_Renderer;

#ifdef DAEDALUS_RELEASE_BUILD
bool g_muxDumpEnabled = false;
#else
bool g_muxDumpEnabled = true;
#endif

#include "combiner.h"

/*
General discussion

The problem of mux to Direct3D translation is that the N64 has a 2-cycle
	structure, where the results of the first cycle can be used everywhere
	in second cycle and even more than one time.

Direct3D instead only provides a "linear" operation model: an operation may
	only use the result of the first operation.

Also, the N64 uses signed values, while DirectX only supports positive values
	(anything < 0 becomes 0)

DirectX 8.0 supports D3DTA_TEMP and allows storing a single value there. This
	would solve a lot of problems, but unfortunately Daedalus uses
	Direct3D 7.0 and the temp register is not supported on most video cards.
	Also, most video cards only support 3 or 4 texture stages in a limited way
	so a 6-stage pass using temp for the firs cycle result is not possible.

Another limitation is that the N64 supports Shade (result of shading),
	Environment (constant), Primitive (constant) and 0 (hardcoded) as
	variables, while Direct3D only supports TFactor (constant), Diffuse
	(result of shading) and Specular.
	However, since constants can be put both into single-color textures and
	variables, this isn't a problem.
	
The tree-based combiner works in the following way:
1. An expression tree is created from the mux value
2. The tree is optimized
3. All the possible combinations are generated and tested by the driver
4. A score is assigned to each combination and the lowest one wins

Limitations:
- Alpha cannot be multipass:
	Alpha must be multiplied with the old value in the first pass, so alpha cannot be multipass
- Some muxes may be incorrect on video cards with less than 3 texture stages
- Complex muxes may be incorrect
- Doesn't take advantage of DirectX 8 temp register and BLENDOP
- Doesn't support signedness (however BLENDxxxALPHA is used when possible, and this partially solves the problem)

TODO
- Support for the second texture
- Optimization for vertex modifications and 0/1 replacement
*/

const CombinerColorComponent::float_type CombinerColorComponent::f0 = (float_type)0;
const CombinerColorComponent::float_type CombinerColorComponent::f1 = (float_type)1;
const CombinerColorComponent::float_type CombinerColorComponent::f2 = (float_type)2;
const CombinerColorComponent::float_type CombinerColorComponent::f0_5 = f1 / f2;
const CombinerColorComponent::float_type CombinerColorComponent::f255 = (float_type)255;
    
vector<const char*> g_combinerErrorMessages;

void NewCombiner_Error(const char* p)
{
	g_combinerErrorMessages.push_back(p);
	DBGConsole_Msg(0, "MUX: %s", p);
}

const char * g_colcombtypes32[32] =
{
	"Combined    ", "Texel0      ",
	"Texel1      ", "Primitive   ", 
	"Shade       ", "Environment ",
	"1           ", "Comb_Alpha  ",
	"Texel0_Alpha", "Texel1_Alpha",
	"Prim_Alpha  ", "Shade_Alpha ",
	"Env_Alpha   ", "LOD_Fraction",
	"PrimLODFrac ", "K5          ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ", "?           ",
	"?           ",	"0           "
};

// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
const char *g_colcombtypes16[16] =
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support
{
	"Combined    ", "Texel0      ",
	"Texel1      ", "Primitive   ", 
	"Shade       ", "Environment ",
	"1           ", "Comb_Alpha  ",
	"Texel0_Alpha", "Texel1_Alpha",
	"Prim_Alpha  ", "Shade_Alpha ",
	"Env_Alpha   ", "LOD_Fraction",
	"PrimLOD_Frac", "0           "
};

// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
const char *g_colcombtypes8[8] =
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support
{
	"Combined    ", "Texel0      ",
	"Texel1      ", "Primitive   ", 
	"Shade       ", "Environment ",
	"1           ", "0           ",
};

const char* g_variableStrings[] =
{
	"Diffuse",
	"Current",
	"Texture",
	"TFactor",
	"Specular",
	
	"PreviousPass",
	
	"Unused",
	"Variable7",
	"0",
	"Primitive",
	"Environment"
};

const char* g_textureOpStrings[] =
{
// Control
	"<ZEROOP>",
    "<disable>",
    "<selectarg1>",
	"<selectarg2>",

// Modulate
    "*",
	"<modulate2x>",
	"<modulate4x>",

// Add
	"+",
	"<addsigned>",
	"<addsigned2x>",
    "-",
    "<addsmooth>",

// Linear alpha blend: Arg1*(Alpha) + Arg2*(1-Alpha)
    "<blenddiffusealpha>",
	"<blendtexturealpha>",
	"<blendfactoralpha>",
	"<blendtexturealphapm>",
	"<blendcurrentalpha>",

// Specular mapping
	"<premodulate>",
	"<modulatealpha_addcolor>",
	"<modulatecolor_addalpha>",
	"<modulateinvalpha_addcolor>",
	"<modulateinvcolor_addalpha>",
    
// Bump mapping
    "<bumpenvmap>",
	"<bumpenvmapluminance>",
    "<dot>",
};

const char* g_textureOpRawStrings[] =
{
// Control
	"D3DTOP 0   ",
    "DISABLE    ",
    "SELECTARG1 ",
	"SELECTARG2 ",

// Modulate
    "MODULATE   ",
	"MODULATE2X ",
	"MODULATE4X ",

// Add
	"ADD        ",
	"ADDSIGNED  ",
	"ADDSIGNED2X",
    "SUBTRACT   ",
    "ADDSMOOTH  ",

// Linear alpha blend: Arg1*(Alpha) + Arg2*(1-Alpha)
    "BLENDDIFFUSEALPHA",
	"BLENDTEXTUREALPHA",
	"BLENDFACTORALPHA",
	"BLENDTEXTUREALPHAPM",
	"BLENDCURRENTALPHA",

// Specular mapping
	"PREMODULATE",
	"MODULATEALPHA_ADDCOLOR",
	"MODULATECOLOR_ADDALPHA",
	"MODULATEINVALPHA_ADDCOLOR",
	"MODULATEINVCOLOR_ADDALPHA",
    
// Bump mapping
    "BUMPENVMAP ",
	"BUMPENVMAPLUMINANCE",
    "DOTPRODUCT3"
};

const char* g_textureArgumentStrings[] =
{
	"Diffuse ",
	"Current ",
	"Texture ",
	"TFactor ",
	"Specular",

	// these should not be used
	"(PreviousPass)",
	
	"(Unused)",
	"(Variable7)",
	"(0)",
	"(Primitive)",
	"(Environment)"
};


//struct CombinerValue
//{
	void CombinerValue::MarkAsInAlpha(vector<CombinerPreliminarStage>& stages)
	{
		for(vector<CombinerPreliminarStage>::iterator iter = stages.begin(); iter < stages.end(); iter++)
		{
			iter->operand1.SetInAlpha();
			iter->operand2.SetInAlpha();
		}
	}

	DWORD CombinerValue::ComputeDepth()
	{
		if(IsUnknownExpression())
		{
			depth = 0;
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
				depth = max((DWORD)depth, iter->ComputeDepth());
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support
			}
			depth++;
		}
		else
		{
			depth = 0;
		}
		return depth;
	}

// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
	CombinerValue CombinerValue::GetValueAndXorFlags(CombinerValueFlags vand, CombinerValueFlags vxor) const
	{
		CombinerValue newValue = *this;
		newValue.flags = (CombinerValueFlags)((newValue.flags & vand) ^ vxor);
		return newValue;
	}
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support

	void CombinerValue::AddConstantSlotToMap(map<CombinerValue, DWORD>& constantSlotMap) const
	{
		if(IsKnownExpression() || IsKnownVariable() || (IsVariable() && ((variable == CombinerVariable_Diffuse) || (variable == CombinerVariable_Specular) || (variable == CombinerVariable_TFactor))))
		{
			// 0 is 0 in all channels, so getalpha doesn't hurt, but allows allocation in alpha slots
			DWORD& usageRef = constantSlotMap[(*this).GetValueAndXorFlags(
				(CombinerValueFlags)~(CombinerValueFlags_Complement | CombinerValueFlags_InAlpha | ((variable == CombinerVariable_0) ? CombinerValueFlags_GetAlpha : 0)),
				(CombinerValueFlags)(0  | ((variable == CombinerVariable_0) ? CombinerValueFlags_GetAlpha : 0))
				)];

			if(IsInAlpha())
				usageRef |= 0x80000000;
			usageRef++;
		}
	}

	void CombinerValue::AddConstantSlotsToMap(const vector<CombinerPreliminarStage>& stages, map<CombinerValue, DWORD>& constantSlotMap)
	{
		for(vector<CombinerPreliminarStage>::const_iterator iter = stages.begin(); iter < stages.end(); iter++)
		{
			if(iter->operation != D3DTOP_SELECTARG2)
				iter->operand1.AddConstantSlotToMap(constantSlotMap);
			if(iter->operation != D3DTOP_SELECTARG1)
				iter->operand2.AddConstantSlotToMap(constantSlotMap);
			if(iter->operation == D3DTOP_BLENDDIFFUSEALPHA)
				CombinerValue(CombinerVariable_Diffuse, true).AddConstantSlotToMap(constantSlotMap);
		}
	}

	void CombinerValue::ApplyConstantSlot(const map<CombinerValue, CombinerValue>& constantSlotMap)
	{
		if(IsKnownExpression() || IsKnownVariable() || (IsVariable() && ((variable == CombinerVariable_Diffuse) || (variable == CombinerVariable_Specular) || (variable == CombinerVariable_TFactor))))
		{
			CombinerValue valueWithoutFlags = (*this).GetValueAndXorFlags(
				(CombinerValueFlags)~(CombinerValueFlags_Complement | CombinerValueFlags_InAlpha | ((variable == CombinerVariable_0) ? CombinerValueFlags_GetAlpha : 0)),
				(CombinerValueFlags)(0  | ((variable == CombinerVariable_0) ? CombinerValueFlags_GetAlpha : 0))
				);
			
			map<CombinerValue, CombinerValue>::const_iterator iter = constantSlotMap.find(valueWithoutFlags);
			if(iter != constantSlotMap.end())
			{
				//NewCombiner_Error("Direct3D variable successfully mapped!");
				CombinerValueFlags oldFlags = flags;
				*this = iter->second;
				flags = (CombinerValueFlags)(flags | (oldFlags & CombinerValueFlags_Complement));
			}
			else
			{
				ostrstream stream;
				Dump(stream);
				NewCombiner_Error("Unmapped non-Direct3D constant");
				NewCombiner_Error(stream.str());
				//*this = CombinerValue(CombinerVariable_Diffuse);
			}
		}
	}

	void CombinerValue::ApplyConstantSlots(vector<CombinerPreliminarStage>& stages, const map<CombinerValue, CombinerValue>& constantSlotMap)
	{
		for(vector<CombinerPreliminarStage>::iterator iter = stages.begin(); iter < stages.end(); iter++)
		{
			if(iter->operation != D3DTOP_SELECTARG2)
				iter->operand1.ApplyConstantSlot(constantSlotMap);
			if(iter->operation != D3DTOP_SELECTARG1)
				iter->operand2.ApplyConstantSlot(constantSlotMap);
		}
	}

	// (a + (b * c)) -> ((b * c) + a)
	void CombinerValue::RearrangeBasedOnDepth()
	{
		if(IsUnknownExpression())
		{
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				iter->RearrangeBasedOnDepth();
			}
			if((expression->operation == D3DTOP_ADD) || (expression->operation == D3DTOP_MODULATE))
				sort<vector<CombinerValue>::iterator>(expression->values.begin(), expression->values.end(), CombinerValue_depth_greater);
		}
	}

	// (a + k1 + k2) -> (a + compute_constant_value(k1 + k2))
	void CombinerValue::OptimizeConstants()
	{
		if(IsExpression())
		{
			DWORD constantCount = 0;
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				iter->OptimizeConstants();
				if(iter->IsKnown())
				{
					constantCount++;
				}
			}
			if(constantCount == expression->values.size())
			{
				// All constants
				SetKnown();
			}
			else if((constantCount >= 2) && ((expression->operation == D3DTOP_MODULATE) || (expression->operation == D3DTOP_ADD) || (expression->operation == D3DTOP_SUBTRACT)))
			{
				CombinerValue constants(expression->operation);
				constants.SetKnown();
				bool firstOperandInConstants = false;
				for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end();)
				{
					if(iter->IsKnown())
					{
						if(iter == expression->values.begin())
							firstOperandInConstants = true;
						constants.expression->values.push_back(*iter);
						iter = expression->values.erase(iter);
					}
					else
					{
						iter++;
					}
				}
				// k1 - a - k2 -> (k1 - k2) - a
				// a - k1 - k2 -> a - (k1 + k2)
				if(expression->operation == D3DTOP_SUBTRACT)
				{
					if(firstOperandInConstants)
					{
						expression->values.insert(expression->values.begin(), constants);
					}
					else
					{
						constants.expression->operation = D3DTOP_ADD;
						expression->values.push_back(constants);
					}
				}
				else
				{
					expression->values.push_back(constants);
				}
			}
		}
	}

	// only call on a binary tree
	void CombinerValue::RemoveNops()
	{
		if(IsUnknownExpression())
		{
			expression->values[0].RemoveNops();
			expression->values[1].RemoveNops();
			bool nop0 = (expression->values[0].IsVariable() && (expression->values[0].variable == CombinerVariable_Nop));
			bool nop1 = (expression->values[1].IsVariable() && (expression->values[1].variable == CombinerVariable_Nop));
			
			if(nop0 || nop1)
				NewCombiner_Error("Unsupported mux operand removed!");

			if(nop0)
			{
				if(nop1)
				{
					*this = CombinerValue(CombinerVariable_0);
				}
				else
				{
					*this = expression->values[1];
				}
			}
			else if(nop1)
			{
				*this = expression->values[0];
			}
		}
	}

	void CombinerValue::AdjustSingleOperandExpression()
	{
		if(IsExpression() && (expression->values.size() == 1))
		{
			bool isComplement = IsComplement();
			bool isGetAlpha = IsGetAlpha();
			*this = *(expression->values.begin());
			if(isComplement)
				Complement();
			if(isGetAlpha)
				SetGetAlpha();
		}
	}

	// (a + 0) -> a, (a * 1) -> a, (a * 0) -> 0, (a - 0) -> a
	void CombinerValue::RemoveRedundantOperations()
	{
		if(IsUnknownExpression())
		{
			if((expression->operation == D3DTOP_ADD) || (expression->operation == D3DTOP_SUBTRACT))
			{
				for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); )
				{
					iter->RemoveRedundantOperations();
					if((*iter == CombinerValue(CombinerVariable_0)) && ((expression->operation != D3DTOP_SUBTRACT) || (iter != expression->values.begin())))
					{
						iter = expression->values.erase(iter);
					}
					else
					{
						iter++;
					}
				}
				
				if(expression->values.size() == 0)
				{
					expression->values.push_back(CombinerValue(CombinerVariable_0));
				}
			}
			if(expression->operation == D3DTOP_SUBTRACT)
			{
				vector<CombinerValue>::iterator iter = expression->values.begin();
				iter++;
				for(; iter < expression->values.end(); iter++)
				{
					if(expression->values[0] == *iter)
					{
						// Saturation allows to ignore other values
						*this = CombinerValue(CombinerVariable_0);
						return; // if we only break, the following code will crash because expression is no longer valid
					}
				}
				if(expression->values[0] == CombinerValue(CombinerVariable_0))
				{
					*this = CombinerValue(CombinerVariable_0);
					return; // if we only break, the following code will crash because expression is no longer valid
				}
				if(expression->values[0] == CombinerValue(CombinerVariable_0, false, true))
				{
					expression->values.erase(expression->values.begin());
					expression->operation = D3DTOP_ADD;
					Complement();
				}
				if(expression->values.size() == 2)
				{
					if(expression->values[0].IsUnknownExpression(D3DTOP_ADD) && expression->values[1].IsUnknownExpression(D3DTOP_ADD))
					{
						for(iter = expression->values[0].expression->values.begin(); iter < expression->values[0].expression->values.end(); )
						{
							vector<CombinerValue>::iterator iter2;
							for(iter2 = expression->values[1].expression->values.begin(); iter2 < expression->values[1].expression->values.end(); iter2++)
							{
								if(*iter == *iter2)
								{
									iter = expression->values[0].expression->values.erase(iter);
									expression->values[1].expression->values.erase(iter2);
									goto skip_iter_inc;
								}
							}
							iter++;
						skip_iter_inc:;
						}

						expression->values[0].AdjustSingleOperandExpression();
						expression->values[1].AdjustSingleOperandExpression();
					}
				}
			}
			if(expression->operation == D3DTOP_MODULATE)
			{
				for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); )
				{
					iter->RemoveRedundantOperations();
					if(*iter == CombinerValue(CombinerVariable_0))
					{
						*this = CombinerValue(CombinerVariable_0);
						return; // if we only break, the following code will crash because expression is no longer valid
						//iter = expression->values.erase(iter);
						//DBGConsole_Msg(0, "[YMultiplication by 0: assuming 1]");
					}
					if(*iter == CombinerValue(CombinerVariable_0, false, true))
					{
						iter = expression->values.erase(iter);
					}
					else
					{
						iter++;
					}
				}
				if(expression->values.size() == 0)
				{
					expression->values.push_back(CombinerValue(CombinerVariable_0, false, true));
				}
			}

			AdjustSingleOperandExpression();
		}
	}

	void CombinerValue::NormalizeGetAlpha()
	{
		if(IsExpression())
		{
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				if(IsGetAlpha())
					iter->SetGetAlpha();

				iter->NormalizeGetAlpha();
			}
			
			flags = (CombinerValueFlags)(flags &~ CombinerValueFlags_GetAlpha);
		}
	}

	void CombinerValue::NormalizeAddSub()
	{
		if(IsUnknownExpression())
		{
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				iter->NormalizeAddSub();
			}
			
			if(expression->operation == D3DTOP_ADD)
			{
				for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
				{
					// a + b + (c - d) -> (a + b + c) - d
					if(iter->IsUnknownExpression() && (iter->expression->operation == D3DTOP_SUBTRACT) && !iter->IsComplement())
					{
						bool isComplement = IsComplement();
						if(isComplement)
							Complement();

						CombinerValue sub = *iter;

						// -> a + b; sub: (b - c)
						*iter = iter->expression->values[0];

						// -> a + b; sub: ((a + b) - c)
						sub.expression->values[0] = *this;

						// ((a + b) - c)
						*this = sub;

						if(isComplement)
							Complement();

						NormalizeAddSub();
						
						break;
					}
				}	
			}
			else if(expression->operation == D3DTOP_SUBTRACT)
			{
				if(expression->values.size() > 2)
				{
					CombinerValue add(D3DTOP_ADD);
					vector<CombinerValue>::iterator iter = expression->values.begin();
					for(; iter < expression->values.end();)
					{
						add.expression->values.push_back(*iter);
						iter = expression->values.erase(iter);
					}
					expression->values.push_back(add);
					NormalizeAddSub();
				}
				// (a - b) - c -> a - (b + c)
				else if(expression->values[0].IsUnknownExpression() && (expression->values[0].expression->operation == D3DTOP_SUBTRACT) && !expression->values[0].IsComplement())
				{
					CombinerValue add(D3DTOP_ADD);
					add.expression->values.push_back(expression->values[0].expression->values[1]);
					add.expression->values.push_back(expression->values[1]);

					expression->values[0] = expression->values[0].expression->values[0];
					expression->values[1] = add;
					NormalizeAddSub();
				}
				// a - (b - c) -> (a + c) - b
				else if(expression->values[1].IsUnknownExpression() && (expression->values[1].expression->operation == D3DTOP_SUBTRACT) && !expression->values[1].IsComplement())
				{
					CombinerValue add(D3DTOP_ADD);
					add.expression->values.push_back(expression->values[0]);
					add.expression->values.push_back(expression->values[1].expression->values[1]);

					expression->values[0] = add;
					expression->values[1] = expression->values[1].expression->values[0];
					NormalizeAddSub();
				}

				// If both are subtractions, the following happens:
				// (a - b) - (a - c) ->
				// a - (b + (a - c) ->
				// a - ((b + a) - c) ->
				// (a + c) - (b + a)
			}
		}
	}

	// ((a + b) + c) -> (a + b + c)
	void CombinerValue::AssociateOperations()
	{
		if(IsUnknownExpression())
		{
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				iter->AssociateOperations();
			}
			// necessary since vector::insert invalidates iterators
			bool continueLoop = true;
			while(continueLoop)
			{
				continueLoop = false;
				if((expression->operation == D3DTOP_MODULATE) || (expression->operation == D3DTOP_ADD))
				{	
					for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
					{
						if(iter->IsUnknownExpression() && (iter->expression->operation == expression->operation) && (!iter->IsComplement()))
						{
							CombinerValue subValue = *iter;
							// a + alpha(b + c) -> a + alpha(b) + alpha(c)
							if(subValue.IsGetAlpha())
							{
								for(vector<CombinerValue>::iterator iter2 = subValue.expression->values.begin(); iter2 < subValue.expression->values.end(); iter2++)
								{
									iter2->SetGetAlpha();
								}
							}
							iter = expression->values.erase(iter);
							expression->values.insert(expression->values.end(), subValue.expression->values.begin(), subValue.expression->values.end());
							continueLoop = true;
							break;
						}
					}
				}
			}
		}
	}

	// undoes association
	void CombinerValue::TransformToBinary()
	{
		if(IsUnknownExpression())
		{
			if(!IsMultiPass())
			{
				if((expression->operation == D3DTOP_SUBTRACT) && (expression->values.size() == 2))
				{
					if(expression->values[1].IsUnknownExpression(D3DTOP_ADD) && (expression->values[1].depth >= 2))
					{
						if(expression->values[0].depth >= 2)
						{
							NewCombiner_Error("TransformToBinary: more than one non-zero depth node under single-pass node");
						}
						
						if(expression->values[0].IsUnknownExpression(D3DTOP_ADD))
							expression->values[0].expression->values[0] = CombinerValue(D3DTOP_SUBTRACT, expression->values[0].expression->values[0], expression->values[1].expression->values[0]);
						else
							expression->values[0] = CombinerValue(D3DTOP_SUBTRACT, expression->values[0], expression->values[1].expression->values[0]);

						expression->values[1].expression->values.erase(expression->values[1].expression->values.begin());
						if(expression->values[1].expression->values.size() == 1)
						{
							if(expression->values[1].IsComplement())
								expression->values[1].expression->values[0].Complement();

							expression->values[1] = expression->values[1].expression->values[0];
						}
					}
					else if(expression->values[1].depth >= 1)
					{
						if(expression->values[0].depth >= 2)
						{
							NewCombiner_Error("TransformToBinary: more than one non-zero depth node under single-pass node");
						}
						
						if(expression->values[0].IsUnknownExpression(D3DTOP_ADD))
							expression->values[0].expression->values[0] = CombinerValue(D3DTOP_SUBTRACT, expression->values[0].expression->values[0], expression->values[1]);
						else
							expression->values[0] = CombinerValue(D3DTOP_SUBTRACT, expression->values[0], expression->values[1]);
						
						if(IsComplement())
							expression->values[0].Complement();

						*this = expression->values[0];
					}
					if(expression->values[1].IsUnknownExpression(D3DTOP_ADD))
					{
						expression->values[1].expression->values.insert(expression->values[1].expression->values.begin(), expression->values[0]);
						expression->values[1].expression->operation = D3DTOP_SUBTRACT;

						if(IsComplement())
							expression->values[1].Complement();

						*this = expression->values[1];
					}
				}
				for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
				{
					iter->TransformToBinary();
				}
				while(expression->values.size() > 2)
				{
					if(expression->values[1].depth != 0)
					{
						NewCombiner_Error("TransformToBinary: more than one non-zero depth node under single-pass node");
					}
					expression->values[1] = CombinerValue(expression->operation, expression->values[0], expression->values[1]);
					expression->values[1].depth = expression->values[0].depth + 1;
					expression->values.erase(expression->values.begin());
				}
			}
		}
	}

	// (a +- k1) * k2 -> a * k2 +- (k1 * k2)
	// (k1 +- a) * k2 -> k1 * k2 +- (a * k2)
	// (a * b) +- (a * c) with (a * b) and (a * c) non-const -> (b +- c) * a
	// (a * b) +- (c * b) with (a * b) and (a * c) non-const -> (a +- c) * b
	void CombinerValue::OptimizeOperations()
	{
		if(IsUnknownExpression())
		{
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				iter->OptimizeOperations();
			}
			if(expression->operation == D3DTOP_MODULATE)
			{
				if(expression->values.size() == 2)
				{
					bool ok = false;
					if(expression->values[0].IsUnknownExpression() && expression->values[1].IsKnown())
					{
						ok = true;
					}
					else if(expression->values[1].IsUnknownExpression() && expression->values[0].IsKnown())
					{
						swap(expression->values[0], expression->values[1]);
						ok = true;
					}
					if(ok)
					{
						CombinerExpression* addExpr = expression->values[0].expression;
						if(((addExpr->operation == D3DTOP_ADD) || (addExpr->operation == D3DTOP_SUBTRACT)) && (addExpr->values.size() == 2))
						{
							if(addExpr->values[0].IsKnown() || addExpr->values[1].IsKnown())
							{
								*this = CombinerValue(addExpr->operation, CombinerValue(D3DTOP_MODULATE, addExpr->values[0], expression->values[1]), CombinerValue(D3DTOP_MODULATE, addExpr->values[1], expression->values[1]));
							}
						}
					}
				}
			}
			else if((expression->operation == D3DTOP_ADD) || (expression->operation == D3DTOP_SUBTRACT))
			{
				if(expression->values.size() == 2)
				{
					// (1 - Diffuse) * a + Diffuse -> (1 - a) * Diffuse + a
					// useful for cards that require a texture in every stage since it avoids having a Current * Diffuse that must become multipass
					UINT diffuseHackOperand = (expression->values[1].IsUnknownExpression() && expression->values[0].IsVariable(CombinerVariable_Diffuse)) ? 1 : 0;

					if((expression->values[0].IsUnknownExpression() && expression->values[1].IsVariable(CombinerVariable_Diffuse)) || diffuseHackOperand)
					{
						if(!expression->values[1 - diffuseHackOperand].IsComplement() && (expression->values[diffuseHackOperand].expression->operation == D3DTOP_MODULATE))
						{
							CombinerValue complementedDiffuse = expression->values[1 - diffuseHackOperand];
							complementedDiffuse.Complement();

							UINT modulateOperand = (expression->values[diffuseHackOperand].expression->values[1] == complementedDiffuse) ? 1 : 0;
							
							if(modulateOperand || (expression->values[diffuseHackOperand].expression->values[0] == complementedDiffuse))
							{
								if(!expression->values[diffuseHackOperand].expression->values[1 - modulateOperand].IsUnknownExpression())
								{
									*this = CombinerValue(D3DTOP_ADD, CombinerValue(D3DTOP_MODULATE, expression->values[diffuseHackOperand].expression->values[1 - modulateOperand], expression->values[1 - diffuseHackOperand]), expression->values[diffuseHackOperand].expression->values[1 - modulateOperand]);
									expression->values[0].expression->values[0].Complement();
								}
							}
						}
					}
				
					if((expression->values.size() == 2) && expression->values[0].IsUnknownExpression() && expression->values[1].IsUnknownExpression())
					{
						CombinerExpression* mulExpr0 = expression->values[0].expression;
						CombinerExpression* mulExpr1 = expression->values[1].expression;
						// TODO: support (a * b * c) +- (a * b * d) with (a * b) and (a * c) non-const -> ((c +- d) * a * b)
						if(
							(mulExpr0->operation == D3DTOP_MODULATE) && (mulExpr1->operation == D3DTOP_MODULATE) &&
							(mulExpr0->values.size() == 2) && (mulExpr1->values.size() == 2)
							)
						{
							bool swap0 = false;
							bool swap1 = false;
							bool equal = false;
							if(mulExpr0->values[0] == mulExpr1->values[1])
							{
								swap1 = true;
							}
							else if(mulExpr0->values[1] == mulExpr1->values[0])
							{
								swap0 = true;
							}
							else if(mulExpr0->values[1] == mulExpr1->values[1])
							{
								swap0 = true;
								swap1 = true;
							}
							else if(mulExpr0->values[0] == mulExpr1->values[0])
							{
								equal = true;
							}
							if(swap0 || swap1 || equal)
							{
								if(swap0)
								{
									swap(mulExpr0->values[0], mulExpr0->values[1]);
								}
								if(swap1)
								{
									swap(mulExpr1->values[0], mulExpr1->values[1]);
								}
								// avoid reversing previous optimization
								if(
									(!mulExpr0->values[0].IsKnown() || !mulExpr0->values[1].IsKnown()) &&
									(!mulExpr1->values[0].IsKnown() || !mulExpr1->values[1].IsKnown())
									)
								{
									*this = CombinerValue(D3DTOP_MODULATE, CombinerValue(expression->operation, mulExpr0->values[1], mulExpr1->values[1]), mulExpr0->values[0]);
								}
							}
						}
					}
				}
			}
		}
	}

	bool CombinerValue::IsMultiPass() const
	{
		return (flags & CombinerValueFlags_MultiPass) ? true : false;
	}
	// a non-multipass node never has a multipass descendent
	//
	bool CombinerValue::MarkMultiPassNodes()
	{
		if(IsUnknownExpression())
		{
			DWORD multiPassChildren = 0;
			DWORD expressionChildren = 0;

			// check for normalized subtractions
			if((expression->operation == D3DTOP_SUBTRACT) && (expression->values.size() == 2))
			{
				for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
				{
					multiPassChildren += iter->MarkMultiPassNodes() ? 1 : 0;
					if(iter->IsUnknownExpression(D3DTOP_ADD))
					{
						for(vector<CombinerValue>::iterator iter2 = iter->expression->values.begin(); iter2 < iter->expression->values.end(); iter2++)
						{
							if(iter2->IsUnknownExpression())
								expressionChildren++;	
						}
					}
					else if(iter->IsUnknownExpression())
						expressionChildren++;
				}
			}
			else
			{
				for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
				{
					multiPassChildren += iter->MarkMultiPassNodes() ? 1 : 0;
					if(iter->IsUnknownExpression())
						expressionChildren++;
				}
			}
			if(multiPassChildren >= 2)
			{
				// (a + b) * (c + d) + (e + f) * (g + h)
				// impossible to do with the current Direct3D architecture without rearranging the expression
				NewCombiner_Error("More than one multipass children in mux tree!");
			}
			if((multiPassChildren >= 1) || (expressionChildren >= 2))
			{
				flags = (CombinerValueFlags)(flags | CombinerValueFlags_MultiPass);
				return true;
			}
			else
			{
				return false;
			}
		}
		return false;
	}

	void CombinerValue::RearrangeTextures()
	{
		// this function is only required if D3DTOP_SUBTRACT is supported
		// TODO: support it in some way
		/*
		if(IsUnknownExpression())
		{
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				iter->RearrangeTextures();
			}
			if(expression->values[1].IsVariable() && (expression->values[1].variable == CombinerVariable_Texture))
			{
				if(expression->operation == D3DTOP_SUBTRACT)
				{
					NewCombiner_Error("Texture in 2nd operand with non-commutative operation!");
				}
			}
		}
		*/
	}

	void CombinerValue::GenerateAllPossibleInfos_MultiPass(CombinerValue* pRootValue, CombinerValue* pAlphaValue, vector<CombinerInfo>& infoVector, vector<CombinerValue>& multiPassVector, vector<CombinerValue>& zeroDepthVector, bool usingDefaultAlpha)
	{
		if(zeroDepthVector.size() > 0)
		{
			//for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter <= expression->values.end(); iter++)
			// an iterator can't be used because *this is overwritten and the iterator would point to the old copy
			for(int i = 0; i <= expression->values.size(); i++)
			{
				CombinerValue backupValue = *this;
				*this = DuplicateTree();	

				vector<CombinerValue> zeroDepthBackupVector;
				zeroDepthBackupVector.resize(zeroDepthVector.size());
				copy(zeroDepthVector.begin(), zeroDepthVector.end(), zeroDepthBackupVector.begin());

				// avoids creating a single-child multipass node that would cause infinite recursion
				if((i != expression->values.size()) && ((expression->values.size() > 1) || (zeroDepthVector.size() > 1)))
				{
					CombinerValue newValue(expression->operation, expression->values[i], zeroDepthVector.back());
					expression->values[i] = newValue;
				}
				else
				{
					expression->values.push_back(zeroDepthVector.back());
				}
				
				zeroDepthVector.pop_back();
				GenerateAllPossibleInfos_MultiPass(pRootValue, pAlphaValue, infoVector, multiPassVector, zeroDepthVector, usingDefaultAlpha);

				zeroDepthVector.resize(zeroDepthBackupVector.size());
				copy(zeroDepthBackupVector.begin(), zeroDepthBackupVector.end(), zeroDepthVector.begin());

				*this = backupValue;
			}
		}
		else
		{
			CombinerValue backupValue = *this;
			*this = DuplicateTree();	

			GenerateAllPossibleInfos(pRootValue, pAlphaValue, infoVector, usingDefaultAlpha, true);

			*this = backupValue;
		}
	}

	void CombinerValue::GenerateAllPossibleInfos(CombinerValue* pRootValue, CombinerValue* pAlphaValue, vector<CombinerInfo>& infoVector, bool usingDefaultAlpha, bool multiPassReordered)
	{
		// we must modify the original value, otherwise changes would not be accessible from pRootValue

		if(!IsUnknownExpression() || !IsMultiPass())
		{
			//block
			{
				CombinerValue backupValue = *this;
				*this = DuplicateTree();

				CombinerInfo info;
				// alpha must always be single-pass
				pRootValue->GenerateInfo(info, pAlphaValue);
				info.score |= (usingDefaultAlpha ? (1 << 24) : 0);
				infoVector.push_back(info);

				*this = backupValue;
			}

			if(IsUnknownExpression())
			{
				CombinerValue backupValue = *this;
				*this = DuplicateTree();

				flags = (CombinerValueFlags)(flags | CombinerValueFlags_MultiPass);
				GenerateAllPossibleInfos(pRootValue, pAlphaValue, infoVector, usingDefaultAlpha);

				*this = backupValue;
			}
		}
		else
		{
			if((expression->operation == D3DTOP_ADD) || (expression->operation == D3DTOP_MODULATE))
			{
				if(!multiPassReordered && (expression->values.size() > 2))
				{
					CombinerValue backupValue = *this;
					*this = DuplicateTree();

					vector<CombinerValue> multiPassVector;
					vector<CombinerValue> zeroDepthVector;

					vector<CombinerValue>::iterator iter;
					//vector<CombinerValue>::iterator iter2;

					for(iter = expression->values.begin(); iter < expression->values.end(); )
					{
						if(iter->IsMultiPass())
						{
							multiPassVector.push_back(*iter);
							iter = expression->values.erase(iter);
						}
						else if(iter->depth == 0)
						{
							zeroDepthVector.push_back(*iter);
							iter = expression->values.erase(iter);
						}
						else
						{
							iter++;
						}
					}

					GenerateAllPossibleInfos_MultiPass(pRootValue, pAlphaValue, infoVector, multiPassVector, zeroDepthVector, usingDefaultAlpha);

					*this = backupValue;
				}
				else
				{
					// this function works because it never happens that more than one children has multiple possible combinations:
					// 1. Stages may be converted to passes only when we are processing the left branch
					// 2. There is never more than one multipass child node

					CombinerValue backupValue = *this;
					*this = DuplicateTree();

					ComputeDepth();
					RearrangeBasedOnDepth();

// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
					for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
					{
						if(iter->IsMultiPass())
						{
							CombinerValue multiPassChildValue = *iter;
							expression->values.erase(iter);
							expression->values.insert(expression->values.begin(), multiPassChildValue);
							break;
						}
					}
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support
					
					expression->values.begin()->GenerateAllPossibleInfos(pRootValue, pAlphaValue, infoVector, usingDefaultAlpha);

					*this = backupValue;
				}					
			}
		}
	}

	void CombinerValue::GenerateBestInfo(CombinerValue* pAlphaValue, CombinerInfo& info)
	{
		vector<CombinerInfo> infoVector;
		GenerateAllPossibleInfos(this, pAlphaValue, infoVector, false);
		CombinerValue defaultAlpha = CombinerValue(D3DTOP_MODULATE, CombinerValue(CombinerVariable_Texture), CombinerValue(CombinerVariable_Diffuse));
		GenerateAllPossibleInfos(this, &defaultAlpha, infoVector, true);
		sort(infoVector.begin(), infoVector.end(), CombinerInfo::score_less);
		//DBGConsole_Msg(0, "Evaluated combiner infos: %u", infoVector.size());
		if(infoVector.size() >= 1)
		{
			info = infoVector[0];
		}
		else
		{
			info.score = 0xffffffff;
		}
		
		if((info.score >> 28) != 0)
		{
			char buffer[256];
			sprintf(buffer, "Mux translation failed - Score: 0x%08X HResult: 0x%08X", info.score, info.hres);
			NewCombiner_Error(buffer);
		}
	}

	void CombinerValue::Process()
	{
		RemoveNops();

		RemoveRedundantOperations();
		NormalizeGetAlpha();
		NormalizeAddSub();
		AssociateOperations();
		RemoveRedundantOperations();
		OptimizeConstants();
		ComputeDepth();
		RearrangeBasedOnDepth();
		
		OptimizeOperations();
		
		// TODO: Is it really necessary to perform all these passes another time?
		RemoveRedundantOperations();
		NormalizeGetAlpha();
		NormalizeAddSub();
		AssociateOperations();
		RemoveRedundantOperations();
		OptimizeConstants();
		ComputeDepth();
		RearrangeBasedOnDepth();

		MarkMultiPassNodes();
		//if(maxTextureStages > 0)
		//	;
	//		OptimizeRenderingPasses(maxTextureStages, pAlphaPass);
	}

	BYTE CombinerValue::ConvertToTextureArgument(bool isAlpha)
	{
		BYTE ret = (IsUnknownExpression() ? D3DTA_CURRENT : (BYTE)variable) | ((IsGetAlpha() && !isAlpha) ? D3DTA_ALPHAREPLICATE : 0) | (IsComplement() ? D3DTA_COMPLEMENT : 0);
		if((ret & 0xf) > D3DTA_SPECULAR)
		{
			NewCombiner_Error("Unexpected non-Direct3D variable");
			NewCombiner_Error(g_variableStrings[ret & 0xf]);
		}
		return ret;
	}

	/*
	Pass order:
	1. Redundant operations removal
	2. Association
	3. Constant Optimization
	4. Depth computation
	5. Depth-based rearranging

	6. Operation Optimization - may screw up all previous passes

	7. Redundant operations removal
	8. Association
	9. Constant Optimization
	10. Depth computation
	11. Depth-based rearranging

	12. MultiPass marking

	13. Rendering pass optimization

	14. Combiner info generation

	15. Alpha/color combining

	*/

	bool Compare(const CombinerValue& v1, const CombinerValue& v2, CombinerValueFlags mask)
	{
		if((v1.flags & mask) != (v2.flags & mask))
			return false;
		
		if((v1.flags & mask) & CombinerValueFlags_Expression)
		{
			return *(v1.expression) == *(v2.expression);
		}
		else if((v1.flags & mask) & CombinerValueFlags_Variable)
		{
			return v1.variable == v2.variable;
		}
		else
		{
			return true;
		}
	}

	void CombinerValue::CheckOperatorConsistency(const CombinerValue& a, const CombinerValue& b)
	{
#ifdef _DEBUG
		if(a.checkingOperators || b.checkingOperators) // avoid infinite recursion
			return;
		a.checkingOperators = true;
		b.checkingOperators = true;
		if((((a == b) ? true : false) ^ ((b == a) ? true : false)))
		{
			NewCombiner_Error("Non-commutative equality operator!");
		}
		if((a < b) && (b < a))
		{
			NewCombiner_Error("Commutative less than operator!");
		}
		if( ((a == b) && (b < a)) || ((a == b) && (a < b)) )
		{
			NewCombiner_Error("Inconsistent equality and less than operators!");
		}
		if(!(a == b) && !(a < b) && !(b < a))
		{
			NewCombiner_Error("All operators false!");
		}
		a.checkingOperators = false;
		b.checkingOperators = false;
#endif
	}

	bool operator ==(const CombinerValue& v1, const CombinerValue& v2)
	{
		CombinerValue::CheckOperatorConsistency(v1, v2);

		if(v1.IsKnownVariable() && v2.IsKnownVariable() && (v1.variable == CombinerVariable_0) && (v2.variable == CombinerVariable_0))
			return
				(v1.flags & (CombinerValueFlags)(CombinerValueFlags_EqualityMask &~ CombinerValueFlags_GetAlpha)) ==
				(v2.flags & (CombinerValueFlags)(CombinerValueFlags_EqualityMask &~ CombinerValueFlags_GetAlpha));
		else
			return Compare(v1, v2, CombinerValueFlags_EqualityMask);
	}
	
	bool operator <(const CombinerValue& v1, const CombinerValue& v2)
	{
		CombinerValue::CheckOperatorConsistency(v1, v2);

		// ensures (!(a == b) || !(a < b)) && (!(a == b) || !(b < a))
		if(v1 == v2)
		{
			return false;
		}

		if(v1.flags > v2.flags)
			return false;

		if(v1.flags < v2.flags)
			return true;
		
		// don't care about the meaning of value
		/*
		if(v1.flags & CombinerValueFlags_Expression)
		{
			return *(v1.expression) < *(v2.expression);
		}
		else if(v1.flags & CombinerValueFlags_Variable)
		{
			return v1.variable < v2.variable;
		}
		else
		{
			return true;
		}
		*/
		return (v1.value < v2.value);
	}

	bool operator >=(const CombinerValue& v1, const CombinerValue& v2)
	{
		return !(v1 < v2);
	}

	bool operator !=(const CombinerValue& v1, const CombinerValue& v2)
	{
		return !(v1 == v2);
	}

	CombinerValue CombinerValue::DuplicateTree() const
	{
		CombinerValue newValue = *this;
		if(newValue.IsExpression())
		{
			CombinerExpression* pNewExpression = new CombinerExpression();
			*pNewExpression = *(newValue.expression);
			for(vector<CombinerValue>::iterator iter = pNewExpression->values.begin(); iter < pNewExpression->values.end(); iter++)
			{
				*iter = iter->DuplicateTree();
			}
			newValue.expression = pNewExpression;
		}
		return newValue;
	}

	CombinerColor CombinerValue::ComputeColor(const CombinerColor& primitive, const CombinerColor& environment)
	{
		CombinerColor color;
		if(IsVariable())
		{
			switch(variable)
			{
				case CombinerVariable_Primitive:
					color = primitive;
					break;
				case CombinerVariable_Environment:
					color = environment;
					break;
				default:
					color = 0;
					break;
			}
		}
		else if(IsExpression())
		{
			vector<CombinerValue>::iterator iter;
			switch(expression->operation)
			{
				case D3DTOP_ADD:
					iter = expression->values.begin();
					color = iter->ComputeColor(primitive, environment);
					for(iter++; iter < expression->values.end(); iter++)
					{
						color += iter->ComputeColor(primitive, environment);
					}
					break;
				
				case D3DTOP_MODULATE:
					iter = expression->values.begin();
					color = iter->ComputeColor(primitive, environment);
					for(iter++; iter < expression->values.end(); iter++)
					{
						color *= iter->ComputeColor(primitive, environment);
					}
					break;

				case D3DTOP_SUBTRACT:
					iter = expression->values.begin();
					color = iter->ComputeColor(primitive, environment);
					for(iter++; iter < expression->values.end(); iter++)
					{
						color -= iter->ComputeColor(primitive, environment);
					}
					break;

				default:
					color = 0;
					break;
			}
		}
		else
		{
			color = 0;
		}
		if(IsGetAlpha())
		{
			color = color.AlphaReplicate();
		}
		if(IsComplement())
		{
			color = ~color;
		}
		return color;
	}

	CombinerColorComponent CombinerValue::ComputeAlpha(const CombinerColor& primitive, const CombinerColor& environment)
	{
		CombinerColorComponent color;
		if(IsVariable())
		{
			switch(variable)
			{
				case CombinerVariable_Primitive:
					color = primitive.a;
					break;
				case CombinerVariable_Environment:
					color = environment.a;
					break;
				default:
					color = 0;
					break;
			}
		}
		else if(IsExpression())
		{
			vector<CombinerValue>::iterator iter;
			switch(expression->operation)
			{
				case D3DTOP_ADD:
					iter = expression->values.begin();
					color = iter->ComputeAlpha(primitive, environment);
					for(iter++; iter < expression->values.end(); iter++)
					{
						color += iter->ComputeAlpha(primitive, environment);
					}
					break;
				
				case D3DTOP_MODULATE:
					iter = expression->values.begin();
					color = iter->ComputeAlpha(primitive, environment);
					for(iter++; iter < expression->values.end(); iter++)
					{
						color *= iter->ComputeAlpha(primitive, environment);
					}
					break;

				case D3DTOP_SUBTRACT:
					iter = expression->values.begin();
					color = iter->ComputeAlpha(primitive, environment);
					for(iter++; iter < expression->values.end(); iter++)
					{
						color -= iter->ComputeAlpha(primitive, environment);
					}
					break;

				default:
					color = 0;
					break;
			}
		}
		else
		{
			color = 0;
		}
		if(IsComplement())
		{
			color = ~color;
		}
		return color;
	}

	// must be binary
	void CombinerValue::TransformSubtractions()
	{
		if(IsUnknownExpression())
		{
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				iter->TransformSubtractions();
			}
			if(expression->operation == D3DTOP_SUBTRACT)
			{
				expression->operation = D3DTOP_ADD;
				Complement();
				expression->values[0].Complement();
			}
		}
	}

	bool CombinerValue::GenerateStages(/*[out]*/ vector<CombinerPreliminarStage>& stages, bool isAlpha, bool useBlendAlpha)
	{
		bool blendAlphaGeneratedGlobal = false;

		CombinerValue dupValue = DuplicateTree();
		CombinerValue* pValue = &dupValue;

		pValue->ComputeDepth();
		pValue->RearrangeBasedOnDepth();
		pValue->TransformToBinary();
		pValue->TransformSubtractions();

		/*
		if(!pValue->IsUnknownExpression())
		{
			*pValue = CombinerValue(D3DTOP_SELECTARG1, *pValue);
		}
		*/

		stages.clear();

		while(pValue != NULL)
		{
			CombinerValue* pNewValue = NULL;

			CombinerPreliminarStage stage;
			
			if(!pValue->IsUnknownExpression())
			{
				stage.operation = D3DTOP_SELECTARG1;
				stage.operand1 = *pValue;
			}
			else
			{
				bool blendAlphaGenerated = false;
				stage.operation = pValue->expression->operation;

				// (b - c) * alpha + c 
				// -> b * alpha + c * (1 - alpha) -> D3DTOP_BLENDxxxALPHA > b : c
				// we also check for known expressions because Direct3D only supports positive values
				if(useBlendAlpha)
				{
					if(pValue->expression->operation == D3DTOP_ADD)
					{
						UINT modulateOperand = 0;
						if(pValue->expression->values[1].IsExpression(D3DTOP_MODULATE))
						{
							modulateOperand = 1;
						}
						if(
							(modulateOperand || pValue->expression->values[0].IsExpression(D3DTOP_MODULATE))
							&& (!pValue->expression->values[modulateOperand].IsComplement())
							)
						{
							// here we check for a transformed subtraction: 1 - ((1 - a) + b)
							UINT subOperand = 0;
							if(pValue->expression->values[modulateOperand].expression->values[1].IsExpression(D3DTOP_ADD))
							{
								subOperand = 1;
							}
							if(
								(subOperand || pValue->expression->values[modulateOperand].expression->values[0].IsExpression(D3DTOP_ADD))
								&& (pValue->expression->values[modulateOperand].expression->values[subOperand].IsComplement())
								)
							{
								UINT positiveOperand = 0;
								if(pValue->expression->values[modulateOperand].expression->values[subOperand].expression->values[1].IsComplement())
								{
									positiveOperand = 1;
								}
								if(positiveOperand || pValue->expression->values[modulateOperand].expression->values[subOperand].expression->values[0].IsComplement())
								{
									if(
										(pValue->expression->values[modulateOperand].expression->values[subOperand].expression->values[1 - positiveOperand] == pValue->expression->values[1 - modulateOperand]) &&
										(pValue->expression->values[modulateOperand].expression->values[1 - subOperand].IsGetAlpha() || isAlpha) &&
										(!pValue->expression->values[modulateOperand].expression->values[1 - subOperand].IsComplement())
										)
									{
										// I hope this uses alpha from the previous color stage and not from the previous alpha stage
										// If this is wrong, just delete/comment the following if block
										if(pValue->expression->values[modulateOperand].expression->values[1 - subOperand].IsVariable(CombinerVariable_Texture))
										{
											stage.operation = D3DTOP_BLENDTEXTUREALPHA;
											blendAlphaGenerated = true;
										}
										else if(pValue->expression->values[modulateOperand].expression->values[1 - subOperand].IsVariable(CombinerVariable_Diffuse))
										{
											stage.operation = D3DTOP_BLENDDIFFUSEALPHA;
											blendAlphaGenerated = true;
										}
										else //(pValue->expression->values[modulateOperand].expression->values[1 - subOperand].IsUnknownExpression())
										{
											pNewValue = &(pValue->expression->values[modulateOperand].expression->values[1 - subOperand]);
											stage.operation = D3DTOP_BLENDCURRENTALPHA;
											blendAlphaGenerated = true;
										}

										if(blendAlphaGenerated)
										{
											stage.operand1 = pValue->expression->values[modulateOperand].expression->values[subOperand].expression->values[positiveOperand];
											stage.operand1.Complement();
											stage.operand2 = pValue->expression->values[modulateOperand].expression->values[subOperand].expression->values[1 - positiveOperand];
										}
									}
								}
							}
						}
					}

					if(!blendAlphaGenerated && (pValue->expression->operation == D3DTOP_ADD))
					{
						UINT subOperand = 0;
						if(pValue->expression->values[1].IsExpression(D3DTOP_ADD))
						{
							subOperand = 1;
						}
						if(
							(subOperand || pValue->expression->values[0].IsExpression(D3DTOP_ADD))
							&& (pValue->expression->values[subOperand].IsComplement())
							)
						{
							UINT positiveOperand = 0;
							if(pValue->expression->values[subOperand].expression->values[1].IsComplement())
							{
								positiveOperand = 1;
							}
							if(positiveOperand || pValue->expression->values[subOperand].expression->values[0].IsComplement())
							{
								if(
									pValue->expression->values[subOperand].expression->values[0].IsExpression(D3DTOP_MODULATE)
									&& pValue->expression->values[subOperand].expression->values[1].IsExpression(D3DTOP_MODULATE)
									&& !pValue->expression->values[subOperand].expression->values[1 - positiveOperand].IsComplement()
									)
								{
									UINT alphaPosOperand = 2;
									UINT alphaNegOperand;
									if(
										pValue->expression->values[subOperand].expression->values[positiveOperand].expression->values[0] == 
										pValue->expression->values[subOperand].expression->values[1 - positiveOperand].expression->values[0]
										)
									{
										alphaPosOperand = 0;
										alphaNegOperand = 0;
									}
									else if(
										pValue->expression->values[subOperand].expression->values[positiveOperand].expression->values[0] == 
										pValue->expression->values[subOperand].expression->values[1 - positiveOperand].expression->values[1]
										)
									{
										alphaPosOperand = 0;
										alphaNegOperand = 1;
									}
									else if(
										pValue->expression->values[subOperand].expression->values[positiveOperand].expression->values[1] == 
										pValue->expression->values[subOperand].expression->values[1 - positiveOperand].expression->values[0]
										)
									{
										alphaPosOperand = 1;
										alphaNegOperand = 0;
									}
									else if(
										pValue->expression->values[subOperand].expression->values[positiveOperand].expression->values[1] == 
										pValue->expression->values[subOperand].expression->values[1 - positiveOperand].expression->values[1]
										)
									{
										alphaPosOperand = 1;
										alphaNegOperand = 1;
									}
									
									if(
										(alphaPosOperand != 2)
										&& (pValue->expression->values[subOperand].expression->values[positiveOperand].expression->values[alphaPosOperand].IsGetAlpha())
										&& (pValue->expression->values[subOperand].expression->values[1 - positiveOperand].expression->values[alphaNegOperand].IsGetAlpha())
										&& (pValue->expression->values[subOperand].expression->values[1 - positiveOperand].expression->values[1 - alphaNegOperand] == pValue->expression->values[1 - subOperand])
										)
									{
										// I hope this uses alpha from the previous color stage and not from the previous alpha stage
										// If this is wrong, just delete/comment the following if block
										if(pValue->expression->values[subOperand].expression->values[positiveOperand].expression->values[alphaPosOperand].IsVariable(CombinerVariable_Texture))
										{
											stage.operation = D3DTOP_BLENDTEXTUREALPHA;
											blendAlphaGenerated = true;
										}
										else if(pValue->expression->values[subOperand].expression->values[positiveOperand].expression->values[alphaPosOperand].IsVariable(CombinerVariable_Diffuse))
										{
											stage.operation = D3DTOP_BLENDDIFFUSEALPHA;
											blendAlphaGenerated = true;
										}
										else //(pValue->expression->values[modulateOperand].expression->values[1 - subOperand].IsUnknownExpression())
										{
											pNewValue = &(pValue->expression->values[subOperand].expression->values[positiveOperand].expression->values[alphaPosOperand]);
											stage.operation = D3DTOP_BLENDCURRENTALPHA;
											blendAlphaGenerated = true;
										}

										if(blendAlphaGenerated)
										{
											stage.operand1 = pValue->expression->values[subOperand].expression->values[positiveOperand].expression->values[1 - alphaPosOperand];
											stage.operand2 = pValue->expression->values[subOperand].expression->values[1 - positiveOperand].expression->values[1 - alphaNegOperand];
										}			
									}
								}
							}
						}
					}
				}

				if(!blendAlphaGenerated)
				{
					if((pValue->expression->values.size() >= 2) && pValue->expression->values[1].IsVariable(CombinerVariable_Texture))
					{
						if((pValue->expression->operation == D3DTOP_ADD) || (pValue->expression->operation == D3DTOP_MODULATE))
						{
							swap(pValue->expression->values[0], pValue->expression->values[1]);
						}
						else if(expression->operation == D3DTOP_SELECTARG2)
						{
							swap(pValue->expression->values[0], pValue->expression->values[1]);
							pValue->expression->operation = D3DTOP_SELECTARG1;
						}
						else if(pValue->expression->operation == D3DTOP_SELECTARG1)
						{
							// ignore
						}
						else if(pValue->expression->operation == D3DTOP_SUBTRACT)
						{
							NewCombiner_Error("Texture in subtraction 2nd operand!");
						}
						else
						{
							NewCombiner_Error("Texture in unsupported operation 2nd operand!");
						}
					}
					
					if(pValue->expression->operation != D3DTOP_SELECTARG2)
					{
						if(pValue->expression->values[0].IsUnknownExpression())
						{
							pNewValue = &(pValue->expression->values[0]);
						}
						//stage.operand1 = pValue->expression->values[0].ConvertToTextureArgument(isAlpha);
						stage.operand1 = pValue->expression->values[0];
					}
					
					if(pValue->expression->operation != D3DTOP_SELECTARG1)
					{
						if(pValue->expression->values[1].IsUnknownExpression())
						{
							if(pNewValue != NULL)
							{
								ostrstream stream;
								dupValue.Dump(stream);
								NewCombiner_Error("More than one children in mux preliminar pass generation!");
								stream << '\0';
								NewCombiner_Error(stream.str());
							}
							pNewValue = &(pValue->expression->values[1]);
						}
						//stage.operand2 = pValue->expression->values[1].ConvertToTextureArgument(isAlpha);
						stage.operand2 = pValue->expression->values[1];
					}
				}

				if(pValue->expression->values.size() > 2)
				{
					NewCombiner_Error("More than 2 operands in mux stages generation");
				}

				if(blendAlphaGenerated)
				{
					blendAlphaGeneratedGlobal = true;
				}
			}

			stages.insert(stages.begin(), stage);

			pValue = pNewValue;
		}

		if(dupValue.IsComplement())
		{
			stages.push_back(CombinerPreliminarStage(D3DTOP_SELECTARG1, CombinerValue(CombinerVariable_Current, false, true)));
		}

		return blendAlphaGeneratedGlobal;
	}

	bool CombinerValue::AssignConstantSlots(map<CombinerValue, DWORD>& constantSlotMap, CombinerValue* constantSlotAssignments, map<CombinerValue, CombinerValue>& constantSlotReplacementMap)
	{
		// 1. Equal constants of getAlpha and inAlpha can be unified as inAlpha
		// 2. getAlpha constants that don't fit can be precomputed and placed in color slots
		//return constantSlotMap.size();

		vector<CombinerValueWithUsageCount> constantSlotVector;

		{
			map<CombinerValue, DWORD>::const_iterator iter = constantSlotMap.begin();
			for(; iter != constantSlotMap.end(); iter++)
			{
				CombinerValueWithUsageCount vuc;
				vuc.value = iter->first;
				vuc.usage = iter->second & 0x7fffffff;
				if(iter->second & 0x80000000)
				{
					vuc.value.SetInAlpha();
				}
				constantSlotVector.push_back(vuc);
			}
		}

		sort(constantSlotVector.begin(), constantSlotVector.end(), CombinerValueWithUsageCount::usage_greater);

		vector<CombinerValue> alphaSlots;
		vector<CombinerValue> colorSlots;

		{
			vector<CombinerValueWithUsageCount>::iterator iter, iter2;

			for(iter = constantSlotVector.begin(); iter < constantSlotVector.end(); )
			{
				if(iter->usage == 0)
				{
					iter = constantSlotVector.erase(iter);
				}
				else
				{
					iter2 = iter;
					iter2++;
					for(; iter2 < constantSlotVector.end(); )
					{
						if(Compare(iter->value, iter2->value, CombinerValueFlags_ConstantSlotEqualityMask))
						{
							iter->usage += iter2->usage;
							iter2 = constantSlotVector.erase(iter2);
						}
						else
						{
							iter2++;
						}
					}
					iter++;
				}
			}

			sort(constantSlotVector.begin(), constantSlotVector.end(), CombinerValueWithUsageCount::usage_greater);

			for(iter = constantSlotVector.begin(); iter < constantSlotVector.end(); )
			{
				if(Compare(iter->value, CombinerValue(CombinerVariable_Diffuse), CombinerValueFlags_ConstantSlotEqualityMask))
				{
					colorSlots.push_back(iter->value);
					iter = constantSlotVector.erase(iter);
				}
				else if(Compare(iter->value, CombinerValue(CombinerVariable_Diffuse, true), CombinerValueFlags_ConstantSlotEqualityMask))
				{
					alphaSlots.push_back(iter->value.GetValueAndXorFlags((CombinerValueFlags)~CombinerValueFlags_InAlpha));
					iter = constantSlotVector.erase(iter);
				}
				else
				{
					iter++;
				}
			}

			for(iter = constantSlotVector.begin(); iter < constantSlotVector.end(); )
			{
				if(iter->value.IsInAlpha())
				{
					if(alphaSlots.size() >= 3)
					{
						NewCombiner_Error("Alpha constant slots overflow in AssignConstantSlots");
					}
					else
					{
						alphaSlots.push_back(iter->value.GetValueAndXorFlags((CombinerValueFlags)~CombinerValueFlags_InAlpha));
					}
					iter = constantSlotVector.erase(iter);
				}
				else
				{
					iter++;
				}
			}

			for(iter = constantSlotVector.begin(); iter < constantSlotVector.end();)
			{
				if(iter->value.IsGetAlpha())
				{
					if(alphaSlots.size() >= 3)
					{
						if(colorSlots.size() >= 3)
						{
							NewCombiner_Error("Alpha and color constant slots overflow in AssignConstantSlots");
						}
						else
						{
							colorSlots.push_back(iter->value);
						}
					}
					else
					{
						alphaSlots.push_back(iter->value);
					}
					iter = constantSlotVector.erase(iter);
				}
				else
				{
					iter++;
				}
			}

			for(iter = constantSlotVector.begin(); iter < constantSlotVector.end(); )
			{
				if(colorSlots.size() >= 3)
				{
					NewCombiner_Error("Color constant slots overflow in AssignConstantSlots");
				}
				else
				{
					colorSlots.push_back(iter->value);
				}
				iter = constantSlotVector.erase(iter);
			}
		}
		
		if(constantSlotAssignments == NULL)
			return true;

		// Obsolete warning (note that this is not necessarily an error) 
		/*
		if((colorSlots.size() == 0) || (alphaSlots.size() == 0))
		{
			NewCombiner_Error("Zero sized slot array");
		}
		*/

		for(int i = 0; i < 6; i++)
		{
			constantSlotAssignments[i] = CombinerValue();
		}

		{
			vector<CombinerValue>::iterator iter;

			for(iter = alphaSlots.begin(); iter < alphaSlots.end(); )
			{
				if(iter->IsUnknownVariable() && (iter->variable == CombinerVariable_Diffuse))
				{
					constantSlotAssignments[CombinerConstantSlot_DiffuseAlpha] = *iter;
					constantSlotReplacementMap[*iter] = CombinerValue(CombinerVariable_Diffuse, true);
					iter = alphaSlots.erase(iter);
				}
				else
				{
					iter++;
				}
			}

			for(iter = colorSlots.begin(); iter < colorSlots.end(); )
			{
				if(iter->IsUnknownVariable() && (iter->variable == CombinerVariable_Diffuse))
				{
					constantSlotAssignments[CombinerConstantSlot_Diffuse] = *iter;
					constantSlotReplacementMap[*iter] = CombinerValue(CombinerVariable_Diffuse);
					iter = colorSlots.erase(iter);
				}
				else
				{
					iter++;
				}
			}

			// Here we no longer need to erase() since the vectors are used for the last time

			for(iter = alphaSlots.begin(); iter < alphaSlots.end(); iter++)
			{
				if((constantSlotAssignments[CombinerConstantSlot_TFactorAlpha] == CombinerValue())) //&& !Compare(iter->value, CombinerValue(CombinerVariable_Diffuse, true), CombinerValueFlags_ConstantSlotEqualityMask))
				{
					constantSlotAssignments[CombinerConstantSlot_TFactorAlpha] = *iter;
					constantSlotReplacementMap[*iter] = CombinerValue(CombinerVariable_TFactor, true);
				}
				else if(constantSlotAssignments[CombinerConstantSlot_DiffuseAlpha] == CombinerValue())
				{
					constantSlotAssignments[CombinerConstantSlot_DiffuseAlpha] = *iter;
					constantSlotReplacementMap[*iter] = CombinerValue(CombinerVariable_Diffuse, true);
				}
				else if(constantSlotAssignments[CombinerConstantSlot_SpecularAlpha] == CombinerValue())
				{
					constantSlotAssignments[CombinerConstantSlot_SpecularAlpha] = *iter;
					constantSlotReplacementMap[*iter] = CombinerValue(CombinerVariable_Specular, true);
				}
			}

			for(iter = colorSlots.begin(); iter < colorSlots.end(); iter++)
			{
				if((constantSlotAssignments[CombinerConstantSlot_TFactor] == CombinerValue())) // && !Compare(iter->value, CombinerValue(CombinerVariable_Diffuse), CombinerValueFlags_ConstantSlotEqualityMask))
				{
					constantSlotAssignments[CombinerConstantSlot_TFactor] = *iter;
					constantSlotReplacementMap[*iter] = CombinerValue(CombinerVariable_TFactor);
				}
				else if(constantSlotAssignments[CombinerConstantSlot_Diffuse] == CombinerValue())
				{
					constantSlotAssignments[CombinerConstantSlot_Diffuse] = *iter;
					constantSlotReplacementMap[*iter] = CombinerValue(CombinerVariable_Diffuse);
				}
				else if(constantSlotAssignments[CombinerConstantSlot_Specular] == CombinerValue())
				{
					constantSlotAssignments[CombinerConstantSlot_Specular] = *iter;
					constantSlotReplacementMap[*iter] = CombinerValue(CombinerVariable_Specular);
				}
			}
		}

		return true;
	}

	void CombinerValue::GenerateInfo(CombinerInfo& info, CombinerValue* pAlphaPass, D3DTEXTUREOP operation)
	{
		if(!IsUnknownExpression())
		{
			*this = CombinerValue(D3DTOP_SELECTARG1, *this); //, CombinerValue(CombinerVariable_Unused));
		}
		if(IsMultiPass())
		{
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				iter->GenerateInfo(info, pAlphaPass, expression->operation);
			}
		}
		else
		{
			CombinerPass pass;
			GenerateBestPass(pAlphaPass, pass, operation, info.passes.size() == 0);
			info.passes.push_back(pass);
		}
		EvaluateInfo(info);
	}

	void CombinerValue::GenerateAllPossiblePasses_TextureConstantSlots(vector<CombinerPass>& passes, CombinerPreliminarPass& pass, int stageNum, bool doAlpha)
	{
		int stagesCount = max(pass.colorStages.size(), pass.alphaStages.size());
		if(stageNum == stagesCount)
		{
			GenerateAllPossiblePasses_Finish(passes, pass);
		}
		else
		{
			vector<CombinerPreliminarStage>& stages = doAlpha ? pass.alphaStages : pass.colorStages;
			vector<CombinerValue>& slots = doAlpha ? pass.alphaTextureSlots : pass.colorTextureSlots;
			if(stageNum < stages.size())
			{
				if(slots[stageNum] == CombinerValue())
				{
					CombinerValue texBackup = slots[stageNum];
					CombinerPreliminarStage stageBackup = stages[stageNum];
					bool textureSlotApplied = false;

					if(stages[stageNum].operand1.IsKnown())
					{
						slots[stageNum] = stages[stageNum].operand1;
						stages[stageNum].operand1 = CombinerValue(CombinerVariable_Texture);
						textureSlotApplied = true;
						
						if(stages[stageNum].operand2.IsKnown())
						{
							NewCombiner_Error(doAlpha ? "Two constants in alpha texture stage!" : "Two constants in color texture stage!");
						}
					}
					else if(stages[stageNum].operand2.IsKnown())
					{
						slots[stageNum] = stages[stageNum].operand2;
						stages[stageNum].operand2 = CombinerValue(CombinerVariable_Texture);
						textureSlotApplied = true;

						if((stages[stageNum].operation == D3DTOP_MODULATE) || (stages[stageNum].operation == D3DTOP_ADD))
						{
							swap(stages[stageNum].operand1, stages[stageNum].operand2);
						}
					}

					if(textureSlotApplied)
					{
						if(!doAlpha)
						{
							GenerateAllPossiblePasses_TextureConstantSlots(passes, pass, stageNum, true);
						}
						else
						{
							GenerateAllPossiblePasses_TextureConstantSlots(passes, pass, stageNum + 1, false);
						}
						
						stages[stageNum] = stageBackup;
						slots[stageNum] = texBackup;
					}
				}
			}

			if(!doAlpha)
			{
				GenerateAllPossiblePasses_TextureConstantSlots(passes, pass, stageNum, true);
			}
			else
			{
				GenerateAllPossiblePasses_TextureConstantSlots(passes, pass, stageNum + 1, false);
			}
		}
	}

	void CombinerValue::GenerateAllPossiblePasses_Shift(vector<CombinerPass>& passes, CombinerPreliminarPass& pass, int stageNum, bool doAlpha)
	{
		vector<CombinerPreliminarStage>& stages = doAlpha ? pass.alphaStages : pass.colorStages;
		int maxShifting = max((int)g_D3DDeviceDesc.wMaxTextureBlendStages - (int)stages.size(), (int)0);
		for(int i = 0; i <= maxShifting; i++)
		{
			vector<CombinerPreliminarStage> backup;
			backup.resize(stages.size());
			copy(stages.begin(), stages.end(), backup.begin());

			//vector<CombinerPreliminarStage>::iterator newIter = iter;
			int newStageNum = stageNum;

			for(int i2 = 0; i2 < i; i2++)
			{
				/*newIter =*/ stages.insert((vector<CombinerPreliminarStage>::iterator)&(stages[stageNum]), CombinerPreliminarStage(D3DTOP_SELECTARG1, CombinerValue(CombinerVariable_Current), CombinerValue()));
				newStageNum++;
			}

			//newIter++;
			newStageNum++;

			// no more shifting can be done
			if((newStageNum >= stages.size()) || (g_D3DDeviceDesc.wMaxTextureBlendStages <= stages.size()))
			{
				if(!doAlpha)
				{
					GenerateAllPossiblePasses_Shift(passes, pass, 0, true);
				}
				else
				{
					int stagesCount = max(pass.colorStages.size(), pass.alphaStages.size());
					pass.alphaTextureSlots.resize(8);
					pass.colorTextureSlots.resize(8);
					
					for(int i = 0; i < stagesCount; i++)
					{
						// TODO: handle multitexturing when it's implemented

						// this sets both alpha and color texture slots since setting only one means creating a new texture of texture0 size and changing its alpha/color value with a constant one; this would be expensive
						if
						(
							(
								(i < pass.colorStages.size())
								&&
								(
									pass.colorStages[i].operand1.IsVariable(CombinerVariable_Texture)
									||
									pass.colorStages[i].operand2.IsVariable(CombinerVariable_Texture)
									||
									(pass.colorStages[i].operation == D3DTOP_BLENDTEXTUREALPHA)
									||
									(pass.colorStages[i].operation == D3DTOP_BLENDTEXTUREALPHAPM)
								)
							)
							||
							(
								(i < pass.alphaStages.size())
								&&
								(
									pass.alphaStages[i].operand1.IsVariable(CombinerVariable_Texture)
									||
									pass.alphaStages[i].operand2.IsVariable(CombinerVariable_Texture)
									||
									(pass.alphaStages[i].operation == D3DTOP_BLENDTEXTUREALPHA) // probably this is invalid
									||
									(pass.alphaStages[i].operation == D3DTOP_BLENDTEXTUREALPHAPM) // probably this is invalid
								)
							)
						)
						{
							pass.colorTextureSlots[i] = CombinerValue(CombinerVariable_Texture);
							pass.alphaTextureSlots[i] = CombinerValue(CombinerVariable_Texture);
						}
						else
						{
							pass.colorTextureSlots[i] = CombinerValue();
							pass.alphaTextureSlots[i] = CombinerValue();
						}
					}

					GenerateAllPossiblePasses_TextureConstantSlots(passes, pass);

					// these must be clear before this is call, so this is safe
					pass.alphaTextureSlots.clear();
					pass.colorTextureSlots.clear();
				}
			}
			else
			{
				GenerateAllPossiblePasses_Shift(passes, pass, newStageNum, doAlpha);
			}

			stages.resize(backup.size());
			copy(backup.begin(), backup.end(), stages.begin());
		}
	}

	void CombinerValue::GenerateAllPossiblePasses_Transform(vector<CombinerPass>& passes, CombinerPreliminarPass& pass, int stageNum, bool doAlpha)
	{
		vector<CombinerPreliminarStage>& stages = doAlpha ? pass.alphaStages : pass.colorStages;
		
		if(stageNum < stages.size())
		{
			if(stages[stageNum].operation == D3DTOP_BLENDTEXTUREALPHA)
			{
				if(stages[stageNum].operand1.IsKnown())
				{
					vector<CombinerPreliminarStage> backup;
					backup.resize(stages.size());
					copy(stages.begin(), stages.end(), backup.begin());
// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
					stages.insert((vector<CombinerPreliminarStage>::iterator)&stages[stageNum], CombinerPreliminarStage(D3DTOP_SELECTARG1, CombinerValue(stages[stageNum].operand1)));
					stages[stageNum].operand1 = CombinerValue(CombinerVariable_Current);
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support
					GenerateAllPossiblePasses_Transform(passes, pass, stageNum, doAlpha);

					stages.resize(backup.size());
					copy(backup.begin(), backup.end(), stages.begin());
				}
				else if(stages[stageNum].operand2.IsKnown())
				{
					vector<CombinerPreliminarStage> backup;
					backup.resize(stages.size());
					copy(stages.begin(), stages.end(), backup.begin());

// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
					stages.insert((vector<CombinerPreliminarStage>::iterator)&stages[stageNum], CombinerPreliminarStage(D3DTOP_SELECTARG1, CombinerValue(stages[stageNum].operand2)));
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support

					stages[stageNum].operand2 = CombinerValue(CombinerVariable_Current);
					GenerateAllPossiblePasses_Transform(passes, pass, stageNum, doAlpha);

					stages.resize(backup.size());
					copy(backup.begin(), backup.end(), stages.begin());
				}
			}
		}

		stageNum++;

		if(stageNum >= stages.size())
		{
			if(!doAlpha)
			{
				GenerateAllPossiblePasses_Transform(passes, pass, 0, true);
			}
			else
			{
				GenerateAllPossiblePasses_Shift(passes, pass, 0, false);
			}
		}
		else
		{
			GenerateAllPossiblePasses_Transform(passes, pass, stageNum, doAlpha);
		}
	}

	void CombinerValue::GenerateAllPossiblePasses(vector<CombinerPass>& passes, CombinerPreliminarPass& pass)
	{
		GenerateAllPossiblePasses_Transform(passes, pass, 0, false);
		//DBGConsole_Msg(0, "Possible passes evaluated: %u", passes.size());
	}

	void CombinerValue::GenerateAllPossiblePasses_Finish(vector<CombinerPass>& passes, CombinerPreliminarPass& pass)
	{
		vector<CombinerPreliminarStage> colorBackup;
		colorBackup.resize(pass.colorStages.size());
		copy(pass.colorStages.begin(), pass.colorStages.end(), colorBackup.begin());
		vector<CombinerPreliminarStage> alphaBackup;
		alphaBackup.resize(pass.alphaStages.size());
		copy(pass.alphaStages.begin(), pass.alphaStages.end(), alphaBackup.begin());

		map<CombinerValue, CombinerValue> constantSlotReplacementMap;

		map<CombinerValue, DWORD> constantSlotMap;

		MarkAsInAlpha(pass.alphaStages);

		AddConstantSlotsToMap(pass.alphaStages, constantSlotMap);
		AddConstantSlotsToMap(pass.colorStages, constantSlotMap);
		
		CombinerValue::AssignConstantSlots(constantSlotMap, pass.constantSlots, constantSlotReplacementMap);

		ApplyConstantSlots(pass.alphaStages, constantSlotReplacementMap);
		ApplyConstantSlots(pass.colorStages, constantSlotReplacementMap);

		CombinerPass fpass;
		pass.ConvertToCombinerPass(fpass);
		EvaluatePass(fpass);

		passes.push_back(fpass);

		pass.colorStages.resize(colorBackup.size());
		copy(colorBackup.begin(), colorBackup.end(), pass.colorStages.begin());
		pass.alphaStages.resize(alphaBackup.size());
		copy(alphaBackup.begin(), alphaBackup.end(), pass.alphaStages.begin());
	}

	// TODO: move this functions to their classes

	// set and return pass score; the best pass has the lowest score
	DWORD CombinerValue::EvaluatePass(CombinerPass& pass)
	{
		CombinerVertexModificationInfo cds; // unused

		int colorStagesCount = 8;
		int alphaStagesCount = 8;
		int i;

		for(i = 0; i < 8; i++)
		{
			if(pass.colorStages[i].operation == D3DTOP_DISABLE)
			{
				colorStagesCount = i;
				break;
			}
		}

		for(i = 0; i < 8; i++)
		{
			if(pass.alphaStages[i].operation == D3DTOP_DISABLE)
			{
				alphaStagesCount = i;
				break;
			}
		}

		int scTexCount = 0;
		
		for(i = 0; i < max(colorStagesCount, alphaStagesCount); i++)
		{
			if(pass.colorTextureSlots[i].IsKnown() || pass.alphaTextureSlots[i].IsKnown())
				scTexCount++;
		}

		HRESULT hres = g_Renderer->ApplyCombinerPass(pass, cds, true);

		// this means that 3, 3 is considered better than 4, 1
		// this also forces failure if the max blending stages are exceeded (safer than trusting ValidateDevice)
		// we reserve bits 16-23 for pass count

		//pass.score = ((FAILED(hres) || (colorStagesCount > g_D3DDeviceDesc.wMaxTextureBlendStages) || (alphaStagesCount > g_D3DDeviceDesc.wMaxTextureBlendStages)) ? 0x10000000 : 0) | ((127 - scTexCount) << 16) | (max(colorStagesCount, alphaStagesCount) << 8) | (colorStagesCount + alphaStagesCount);
		//pass.score = (FAILED(hres) ? 0x10000000 : 0) | (scTexCount << 16) | (max(colorStagesCount, alphaStagesCount) << 8) | (colorStagesCount + alphaStagesCount);
		pass.score = ((FAILED(hres) || (colorStagesCount > g_D3DDeviceDesc.wMaxTextureBlendStages) || (alphaStagesCount > g_D3DDeviceDesc.wMaxTextureBlendStages)) ? 0x10000000 : 0) | (scTexCount << 16) | (max(colorStagesCount, alphaStagesCount) << 8) | (colorStagesCount + alphaStagesCount);
		pass.hres = hres;

		if((pass.score >> 28) > 0)
		{
			hres = hres; // dummy for debugging
		}
		return pass.score;
	}

	/*
	28-31: Number of ValidateDevice errors
	27: Default alpha
	23-26: Number of passes - 1
	16-22: Number of constant textures
	8-15:  Sum of number of stages used for either color and alpha from all passes
	0-7:  Sum of number of stages used for color and stages used for alpha from all passes
	The lowest score is the best
	This assumes that a combination is always better than another with more passes but a substantial lower number of stages
	*/

	DWORD CombinerValue::EvaluateInfo(CombinerInfo& info)
	{
		info.score = 0;
		info.hres = S_OK;

		for(vector<CombinerPass>::const_iterator iter = info.passes.begin(); iter < info.passes.end(); iter++)
		{
			info.score += iter->score;
			if(FAILED(iter->hres))
				info.hres = iter->hres;
		}

		// the commented line gives priority to the info with the greatest number of passes - sometimes useful when debugging
		//info.score |= ((15 - info.passes.size()) << 23);
		info.score |= ((info.passes.size() - 1) << 23);
		return info.score;
	}

	void CombinerValue::GenerateBestPass(CombinerValue* pAlphaPass, CombinerPass& outPass, D3DTEXTUREOP operation, bool isFirst)
	{	
		vector<CombinerPass> passes;

		// if true, BLENDxxxALPHA was generated for the channel, so we must also try to generate stages without using BLENDxxxALPHA
		bool transformedColor;
		bool transformedAlpha;
		
		{
			CombinerPreliminarPass pass;
			transformedColor = GenerateStages(pass.colorStages, false, true);
			if(pAlphaPass)
				transformedAlpha = pAlphaPass->GenerateStages(pass.alphaStages, true, true);
			else
				transformedAlpha = false;
			GenerateAllPossiblePasses(passes, pass);
		}

		if(transformedColor)
		{
			//block
			{
				CombinerPreliminarPass pass;
				GenerateStages(pass.colorStages, false, false);
				if(pAlphaPass)
					pAlphaPass->GenerateStages(pass.alphaStages, true, true);
				GenerateAllPossiblePasses(passes, pass);
			}

			if(transformedAlpha)
			{
				//block
				{
					CombinerPreliminarPass pass;
					GenerateStages(pass.colorStages, false, false);
					if(pAlphaPass)
						pAlphaPass->GenerateStages(pass.alphaStages, true, false);
					GenerateAllPossiblePasses(passes, pass);
				}

				//block
				{
					CombinerPreliminarPass pass;
					GenerateStages(pass.colorStages, false, true);
					if(pAlphaPass)
						pAlphaPass->GenerateStages(pass.alphaStages, true, false);
					GenerateAllPossiblePasses(passes, pass);
				}
			}
		}
		else if(transformedAlpha)
		{
			CombinerPreliminarPass pass;
			GenerateStages(pass.colorStages, false, true);
			if(pAlphaPass)
				pAlphaPass->GenerateStages(pass.alphaStages, true, false);
			GenerateAllPossiblePasses(passes, pass);
		}

		sort(passes.begin(), passes.end(), CombinerPass::score_less);

		outPass = *passes.begin();
		
		//DBGConsole_Msg(0, "Combiner pass score: %08X", outPass.score);

		if(pAlphaPass != NULL)
		{
			if(isFirst || (operation == D3DTOP_ADD))
			{
				// ADD: original * (1 - alpha) + (pass1 + pass2) * alpha -> (original * (1 - alpha) + pass1 * alpha) * 1 + pass2 * alpha
				outPass.srcBlend = D3DBLEND_SRCALPHA;
				outPass.destBlend = isFirst ? D3DBLEND_INVSRCALPHA : D3DBLEND_ONE;
			}
			else if(operation == D3DTOP_MODULATE)
			{
				// MODULATE: original * (1 - alpha) + pass1 * pass2 * alpha -> (original * (1 - alpha) + pass1 * alpha) * 0 + pass2 * <DESTCOLOR> (original * (1 - alpha) + pass1 * alpha)
				// TODO: is this correct?
				outPass.srcBlend = isFirst ? D3DBLEND_SRCALPHA : D3DBLEND_DESTCOLOR;
				outPass.destBlend = isFirst ? D3DBLEND_INVSRCALPHA : D3DBLEND_ZERO;
			}
			else
			{
				NewCombiner_Error("Unsupported operation in preliminar pass");
				outPass.srcBlend = D3DBLEND_SRCALPHA;
				outPass.destBlend = isFirst ? D3DBLEND_INVSRCALPHA : D3DBLEND_ONE;
			}
			
			outPass.blend = true;
			outPass.alphaTest = true;
		}
		else
		{
			// TODO: is there a better way to disable alpha?
			outPass.alphaStages[0].operation = D3DTOP_SELECTARG1;
			outPass.alphaStages[0].operand1 = D3DTA_TFACTOR; // don't care about the value - this should be the cheapest

			if(isFirst || (operation == D3DTOP_ADD))
			{
				// ADD: original * 0 + pass1 + pass2 -> (original * 0 + pass1 * 1) * 1 + pass2 * 1
				outPass.srcBlend = D3DBLEND_ONE;
				outPass.destBlend = isFirst ? D3DBLEND_ZERO : D3DBLEND_ONE;
			}
			else if(operation == D3DTOP_MODULATE)
			{
				// MODULATE: original * 0 + pass1 * pass2 -> (original * 0 + pass1 * 1) * 0 + pass2 * <DESTCOLOR> (original * 0 + pass1 * 1)
				outPass.srcBlend = D3DBLEND_DESTCOLOR;
				outPass.destBlend = D3DBLEND_ZERO;
			}
			else
			{
				NewCombiner_Error("Unsupported operation in preliminar pass");
				outPass.srcBlend = D3DBLEND_ONE;
				outPass.destBlend = isFirst ? D3DBLEND_ZERO : D3DBLEND_ONE;
			}

			outPass.blend = !isFirst;
			outPass.alphaTest = false;
		}
	}

	void CombinerValue::MarkAsInUse(bool mark)
	{
		if(IsExpression())
		{
			expression->MarkAsInUse(mark);
			for(vector<CombinerValue>::iterator iter = expression->values.begin(); iter < expression->values.end(); iter++)
			{
				iter->MarkAsInUse(mark);
			}
		}
	}
//};

/*************************************************************************************************************************************************************/

//struct CombinerStage
//{
// BEGIN MODIFIED BY Lkb - 12/jul/2001 - GCC support
	void CombinerStage::DumpOperand(std::ostream& stream, BYTE operand) const
	{
		stream << ((operand & D3DTA_ALPHAREPLICATE) ? "A(" : "  ") << ((operand & D3DTA_COMPLEMENT) ? "1 - " : "    ") << g_textureArgumentStrings[operand & 0x7] << ((operand & D3DTA_ALPHAREPLICATE) ? ")" : " ");
	}

	void CombinerStage::Dump(std::ostream& stream) const
	{
		stream << g_textureOpRawStrings[operation] << " > ";
		DumpOperand(stream, operand1);
		stream << " : ";
		DumpOperand(stream, operand2);
	}
// END MODIFIED BY Lkb - 12/jul/2001 - GCC support
//};

/*************************************************************************************************************************************************************/

//struct CombinerPreliminarStage
//{
	void CombinerPreliminarStage::ConvertToCombinerStage(CombinerStage& stage, bool isAlpha)
	{
		stage.operation = operation;
		stage.operand1 = operand1.ConvertToTextureArgument(isAlpha);
		stage.operand2 = operand2.ConvertToTextureArgument(isAlpha);
	}

//};

/*************************************************************************************************************************************************************/

//struct CombinerPreliminarPass
//{
	void CombinerPreliminarPass::ConvertToCombinerPass(CombinerPass& pass)
	{
		for(int i = 0; i < 8; i++)
		{
			pass.colorTextureSlots[i] = CombinerValue();
			pass.alphaTextureSlots[i] = CombinerValue();

			if(i < colorStages.size())
				colorStages[i].ConvertToCombinerStage(pass.colorStages[i], false);
			else
				pass.colorStages[i] = CombinerStage();

			if(i < alphaStages.size())
				alphaStages[i].ConvertToCombinerStage(pass.alphaStages[i], true);
			else
				pass.alphaStages[i] = CombinerStage();
		}
		copy(constantSlots, constantSlots + 6, pass.constantSlots);
		copy(colorTextureSlots.begin(), colorTextureSlots.end(), pass.colorTextureSlots);
		copy(alphaTextureSlots.begin(), alphaTextureSlots.end(), pass.alphaTextureSlots);

		pass.srcBlend = srcBlend;
		pass.destBlend = destBlend;
		pass.blend = blend;
		pass.alphaTest = alphaTest;
	}
//}

/*************************************************************************************************************************************************************/

//struct CombinerExpression
//{
	CombinerExpression::CombinerExpression(D3DTEXTUREOP p_operation, CombinerValue p_left, CombinerValue p_right)
			: operation(p_operation)
	{
		values.push_back(p_left);
		values.push_back(p_right);
	}

	CombinerExpression::CombinerExpression(D3DTEXTUREOP p_operation, CombinerValue p_operand)
			: operation(p_operation)
	{
		values.push_back(p_operand);
	}

	bool operator ==(const CombinerExpression& v1, const CombinerExpression& v2)
	{
		if((v1.operation != v2.operation) || (v1.values.size() != v2.values.size()))
			return false;
		vector<CombinerValue>::const_iterator iter = v1.values.begin();
		vector<CombinerValue>::const_iterator iter2 = v2.values.begin();
		for(; iter < v1.values.end(); iter++, iter2++)
		{
			if(*iter != *iter2)
				return false;
		}
		return true;
	}

	bool operator <(const CombinerExpression& v1, const CombinerExpression& v2)
	{
		if((v1.operation >= v2.operation) || (v1.values.size() >= v2.values.size()))
			return false;
		vector<CombinerValue>::const_iterator iter = v1.values.begin();
		vector<CombinerValue>::const_iterator iter2 = v2.values.begin();
		for(; iter < v1.values.end(); iter++, iter2++)
		{
			if((*iter) >= (*iter2))
				return false;
		}
		return true;
	}
//};

/*************************************************************************************************************************************************************/

const char* g_d3dblendStrings[] =
{
	"(D3DBLEND)0",
	"D3DBLEND_ZERO",
	"D3DBLEND_ONE",
	"D3DBLEND_SRCCOLOR",
	"D3DBLEND_INVSRCCOLOR",
	"D3DBLEND_SRCALPHA",
	"D3DBLEND_INVSRCALPHA",
	"D3DBLEND_DESTALPHA",
	"D3DBLEND_INVDESTALPHA",
	"D3DBLEND_DESTCOLOR",
	"D3DBLEND_INVDESTCOLOR",
	"D3DBLEND_SRCALPHASAT",
	"D3DBLEND_BOTHSRCALPHA",
	"D3DBLEND_BOTHINVSRCALPHA"
};

const char* g_constantSlotStrings[] =
{
	"TFactor",
	"Diffuse",
	"Specular",
	"TFactor Alpha",
	"Diffuse Alpha",
	"Specular Alpha"
};

const MuxValue g_Mux32[32] = 
{
	MUX_COMBINED, MUX_TEXEL0,   MUX_TEXEL1, MUX_PRIM,
	MUX_SHADE,    MUX_ENV,      MUX_1,      MUX_COMBALPHA,
	MUX_T0_ALPHA, MUX_T1_ALPHA, MUX_PRIM_ALPHA, MUX_SHADE_ALPHA,
	MUX_ENV_ALPHA, MUX_LODFRAC, MUX_PRIMLODFRAC, MUX_UNK,		// Actually k5
	MUX_UNK, MUX_UNK, MUX_UNK, MUX_UNK,
	MUX_UNK, MUX_UNK, MUX_UNK, MUX_UNK,
	MUX_UNK, MUX_UNK, MUX_UNK, MUX_UNK,
	MUX_UNK, MUX_UNK, MUX_UNK, MUX_0
};

const MuxValue g_Mux16[16] = 
{
	MUX_COMBINED, MUX_TEXEL0,   MUX_TEXEL1, MUX_PRIM,
	MUX_SHADE,    MUX_ENV,      MUX_1,      MUX_COMBALPHA,
	MUX_T0_ALPHA, MUX_T1_ALPHA, MUX_PRIM_ALPHA, MUX_SHADE_ALPHA,
	MUX_ENV_ALPHA, MUX_LODFRAC, MUX_PRIMLODFRAC, MUX_0	
};

const MuxValue g_Mux8[8] = 
{
	MUX_COMBINED, MUX_TEXEL0,   MUX_TEXEL1, MUX_PRIM,
	MUX_SHADE,    MUX_ENV,      MUX_1,      MUX_0
};

enum MuxRawValue
{
	MuxRawValue_Combined = 0,
	MuxRawValue_Texel0 = 1,
	MuxRawValue_Texel1 = 2,
	MuxRawValue_Primitive = 3,
	MuxRawValue_Shade = 4,
	MuxRawValue_Environment = 5,
	MuxRawValue_1 = 6,
	MuxRawValue_8_0 = 7,
	MuxRawValue_CombinedAlpha = 7,
	MuxRawValue_Texel0Alpha = 8,
	MuxRawValue_Texel1Alpha = 9,
	MuxRawValue_PrimitiveAlpha = 10,
	MuxRawValue_ShadeAlpha = 11,
	MuxRawValue_EnvironmentAlpha = 12,
	MuxRawValue_LODFrac = 13,
	MuxRawValue_PrimLODFrac = 14,
	MuxRawValue_16_0 = 15,
	MuxRawValue_32_0 = 31
};

const CombinerValue g_MuxValueToCombinerValue[] =
{
	CombinerValue(CombinerVariable_0),
	CombinerValue(CombinerVariable_0, false, true),
	CombinerValue(CombinerVariable_PreviousPass),
	CombinerValue(CombinerVariable_Texture),
	CombinerValue(CombinerVariable_Nop), // TODO: Texel1
	CombinerValue(CombinerVariable_Primitive),
	CombinerValue(CombinerVariable_Diffuse),
	CombinerValue(CombinerVariable_Environment),
	CombinerValue(CombinerVariable_PreviousPass, true),
	CombinerValue(CombinerVariable_Texture, true),
	CombinerValue(CombinerVariable_Nop), // TODO: Texel1
	CombinerValue(CombinerVariable_Primitive, true),
	CombinerValue(CombinerVariable_Diffuse, true),
	CombinerValue(CombinerVariable_Environment, true),
	CombinerValue(CombinerVariable_Nop), // TODO: LODFrac
	CombinerValue(CombinerVariable_Nop), // TODO: PrimLODFrac
	CombinerValue(CombinerVariable_Nop) // Unk
};

CombinerValue ProcessCombinerValue(CombinerValue value)
{
	return value;
}

CombinerValue CombinerValueFromMuxValue(MuxValue muxValue)
{
	return ProcessCombinerValue(g_MuxValueToCombinerValue[muxValue]);
}

CombinerValue CombinerValueFromMux8(DWORD muxValue)
{
	return CombinerValueFromMuxValue(g_Mux8[muxValue & 0x7]);
}

CombinerValue CombinerValueFromMux8InAlpha(DWORD muxValue)
{
	CombinerValue value = CombinerValueFromMuxValue(g_Mux8[muxValue & 0x7]);
	value.SetInAlpha();
	return value;
}

CombinerValue CombinerValueFromMux16(DWORD muxValue)
{
	return CombinerValueFromMuxValue(g_Mux16[muxValue & 0xf]);
}

CombinerValue CombinerValueFromMux32(DWORD muxValue)
{
	return CombinerValueFromMuxValue(g_Mux32[muxValue & 0x1f]);
}

D3DXCOLOR ColorByteToFloat(D3DCOLOR byteColor)
{
	D3DXCOLOR floatColor;
	floatColor.r = ((float)RGBA_GETRED(byteColor)) / 255.0f;
	floatColor.g = ((float)RGBA_GETGREEN(byteColor)) / 255.0f;
	floatColor.b = ((float)RGBA_GETBLUE(byteColor)) / 255.0f;
	floatColor.a = ((float)RGBA_GETALPHA(byteColor)) / 255.0f;
	return floatColor;
}

D3DCOLOR ColorFloatToByte(D3DXCOLOR& color)
{
	// BEGIN MODIFIED BY Lkb - 9/jun/2001 - GCC support
	return RGBA_MAKE(
		(BYTE)(float)(color.r * 255.0f),
		(BYTE)(float)(color.g * 255.0f),
		(BYTE)(float)(color.b * 255.0f),
		(BYTE)(float)(color.a * 255.0f)
		);
	// END MODIFIED BY Lkb - 9/jun/2001 - GCC support
}

static ofstream dumpFile;
static bool dumpFileInit = false;

map<u64, CombinerInfo> muxMap;

void PerformCombinerGarbageCollection(CombinerInfo& info)
{
	info.MarkAsInUse(true);
	vector<CombinerExpression*>::iterator iter;
	for(iter = allocatedExpressions.begin(); iter < allocatedExpressions.end(); iter++)
	{
		if((*iter)->IsInUse())
		{
			(*iter)->MarkAsInUse(false);
		}
		else
		{
			delete (*iter);
		}
	}
	allocatedExpressions.clear();
}

CombinerInfo& ComputeCombinerInfoFromMux(u64 mux, CombinerInfo& info)
{
	info.mux = mux;

	DBGConsole_Msg(0, "MUX: Mux 0x%016I64X used for the first time!", mux);

	if(1)
	{
		DWORD dwMux0 = (DWORD)(mux>>32);
		DWORD dwMux1 = (DWORD)(mux);
		
		CombinerValue values[2][2][4];

		values[0][0][0] = CombinerValueFromMux16(dwMux0>>20);
		values[0][0][1] = CombinerValueFromMux16(dwMux1>>28);
		values[0][0][2] = CombinerValueFromMux32(dwMux0>>15);
		values[0][0][3] = CombinerValueFromMux8(dwMux1>>15);

		values[0][1][0] = CombinerValueFromMux16(dwMux0>>5);
		values[0][1][1] = CombinerValueFromMux16(dwMux1>>24);
		values[0][1][2] = CombinerValueFromMux32(dwMux0);
		values[0][1][3] = CombinerValueFromMux8(dwMux1>>6);

		values[1][0][0] = CombinerValueFromMux8InAlpha(dwMux0>>12);
		values[1][0][1] = CombinerValueFromMux8InAlpha(dwMux1>>12);
		values[1][0][2] = CombinerValueFromMux8InAlpha(dwMux0>>9);
		values[1][0][3] = CombinerValueFromMux8InAlpha(dwMux1>>9);

		values[1][1][0] = CombinerValueFromMux8InAlpha(dwMux1>>21);
		values[1][1][1] = CombinerValueFromMux8InAlpha(dwMux1>>3);
		values[1][1][2] = CombinerValueFromMux8InAlpha(dwMux1>>18);
		values[1][1][3] = CombinerValueFromMux8InAlpha(dwMux1);

		CombinerValue passes[2][2];

		for(int iPassType = 0; iPassType < 2; iPassType++) // 0 = color 1 = alpha
		{
			for(int iPass = 0; iPass < 2; iPass++)
			{
				for(int iValue = 0; iValue < 4; iValue++)
				{
					if(values[iPassType][iPass][iValue] == CombinerValue(CombinerVariable_PreviousPass))
					{
						if(iPass == 0)
						{
							NewCombiner_Error("\"Combined\" in first cycle");
							values[iPassType][iPass][iValue] = CombinerValue(CombinerVariable_0);
						}
						else
						{
							values[iPassType][iPass][iValue] = passes[iPassType][iPass - 1];
						}
					}
					else if(values[iPassType][iPass][iValue] == CombinerValue(CombinerVariable_PreviousPass, true))
					{
						if(iPass == 0)
						{
							NewCombiner_Error("\"Combined\" in first cycle");
							values[iPassType][iPass][iValue] = CombinerValue(CombinerVariable_0);
						}
						else
						{
							values[iPassType][iPass][iValue] = passes[iPassType][iPass - 1];
							values[iPassType][iPass][iValue].SetGetAlpha();
						}
					}
				}
				passes[iPassType][iPass] = CombinerValue(D3DTOP_ADD, CombinerValue(D3DTOP_MODULATE, CombinerValue(D3DTOP_SUBTRACT, values[iPassType][iPass][0], values[iPassType][iPass][1]), values[iPassType][iPass][2]), values[iPassType][iPass][3]);	
			}
		}

		passes[1][1].Process();

		if(passes[1][1] == CombinerValue(CombinerVariable_0))
		{
			// alpha == 0 -> transparent: the primitive must be skipped and this is achieved by emitting no passes
		}
		else
		{
			bool alpha = (passes[1][1] != CombinerValue(CombinerVariable_0, false, true));
			// TODO: handle multipass alpha - currently it is always generated as a single pass, and if it's not possible the combiner may fail
			
			passes[0][1].Process();

			passes[0][1].GenerateBestInfo(alpha ? &(passes[1][1]) : NULL, info);
		}
	}

	info.mux = mux;

	// OK, now, finally, we finished!
	PerformCombinerGarbageCollection(info);

	if(g_muxDumpEnabled)
	{
		if(!dumpFileInit)
		{
			TCHAR dumpFileName[MAX_PATH];
			PathCombine(dumpFileName, g_szDaedalusExeDir, TEXT("daedalus_combiner.txt"));
			dumpFile.open(dumpFileName);
			dumpFileInit = true;
		}
	
		info.Dump(dumpFile);
	}

	return info;
}

void ComputeCombinerInfoFromMuxAndDiscard(u64 mux)
{
	CombinerInfo info;
	ComputeCombinerInfoFromMux(mux, info);
}

CombinerInfo& D3DRender::FindOrComputeCombinerInfoFromMux(u64 mux)
{
	map<u64, CombinerInfo>::iterator combineModeIter = muxMap.find(mux);
	if(combineModeIter != muxMap.end())
		return combineModeIter->second;

	return ComputeCombinerInfoFromMux(mux, muxMap[mux]);
}

extern LPDIRECTDRAWSURFACE7 GetSingleColorTexture(D3DCOLOR color);

HRESULT D3DRender::ApplyCombinerPass(CombinerPass& pass, CombinerVertexModificationInfo& cds, bool test)
{
	m_pD3DDev->SetRenderState( D3DRENDERSTATE_ALPHABLENDENABLE, pass.blend);
	m_pD3DDev->SetRenderState( D3DRENDERSTATE_ALPHATESTENABLE, pass.alphaTest);
	m_pD3DDev->SetRenderState( D3DRENDERSTATE_SRCBLEND, pass.srcBlend);
	m_pD3DDev->SetRenderState( D3DRENDERSTATE_DESTBLEND, pass.destBlend);

	int iStage;
	for(iStage = 0; iStage < 8; iStage++)
	{
		SetColorOp(iStage, pass.colorStages[iStage].operation);
		//if(pass.colorStages[iStage].operation == D3DTOP_DISABLE)
		//	break;
		SetColorArg1(iStage, pass.colorStages[iStage].operand1);
		SetColorArg2(iStage, pass.colorStages[iStage].operand2);
	}

	for(iStage = 0; iStage < 8; iStage++)
	{
		SetAlphaOp(iStage, pass.alphaStages[iStage].operation);
		//if(pass.alphaStages[iStage].operation == D3DTOP_DISABLE)
		//	break;
		SetAlphaArg1(iStage, pass.alphaStages[iStage].operand1);
		SetAlphaArg2(iStage, pass.alphaStages[iStage].operand2);
	}

	CombinerColor primitive((D3DCOLOR)m_dwPrimitiveColor);
	CombinerColor environment((D3DCOLOR)m_dwEnvColor);

	cds.texturesCount = 0;

	for(iStage = 0; iStage < 8; iStage++)
	{
		LPDIRECTDRAWSURFACE7 texture = NULL;
		if(pass.colorTextureSlots[iStage].IsVariable(CombinerVariable_Texture) || pass.alphaTextureSlots[iStage].IsVariable(CombinerVariable_Texture))
		{
			texture = m_lpsTexture;
			cds.textureTypes[iStage] = TextureType_Texture0;
		}
		else
		{
			//*
			if(pass.colorTextureSlots[iStage] != CombinerValue())
			{
				if(pass.alphaTextureSlots[iStage] != CombinerValue())
				{
					texture = GetSingleColorTexture((D3DCOLOR)MergeColorAndAlpha(pass.colorTextureSlots[iStage].ComputeColor(primitive, environment), pass.alphaTextureSlots[iStage].ComputeAlpha(primitive, environment)));
					cds.textureTypes[iStage] = TextureType_SingleColor;
				}
				else
				{
					texture = GetSingleColorTexture((D3DCOLOR)pass.colorTextureSlots[iStage].ComputeColor(primitive, environment));
					cds.textureTypes[iStage] = TextureType_SingleColor;
				}
			}
			else
			{
				if(pass.alphaTextureSlots[iStage] != CombinerValue())
				{
					texture = GetSingleColorTexture((D3DCOLOR)CombinerColor(pass.alphaTextureSlots[iStage].ComputeAlpha(primitive, environment)));
					cds.textureTypes[iStage] = TextureType_SingleColor;
				}
				else
				{
					cds.textureTypes[iStage] = TextureType_Unused;
				}
			}
			//*/
		}

		if(texture != NULL)
		{
			cds.texturesCount = iStage + 1;
		}

		m_pD3DDev->SetTexture( iStage,  texture );
	}

	if(test)
	{
		DWORD passesNum; // what's the purpose of passesNum?
		return m_pD3DDev->ValidateDevice(&passesNum);
	}
	
	SetTextureFactor((D3DCOLOR)MergeColorAndAlpha(pass.constantSlots[CombinerConstantSlot_TFactor].ComputeColor(primitive, environment), pass.constantSlots[CombinerConstantSlot_TFactorAlpha].ComputeAlpha(primitive, environment)));

	cds.overrideDiffuseColor = ((pass.constantSlots[CombinerConstantSlot_Diffuse] != CombinerValue()) && (pass.constantSlots[CombinerConstantSlot_Diffuse] != CombinerValue(CombinerVariable_Diffuse)));
	cds.overrideDiffuseAlpha = ((pass.constantSlots[CombinerConstantSlot_DiffuseAlpha] != CombinerValue()) && (pass.constantSlots[CombinerConstantSlot_DiffuseAlpha] != CombinerValue(CombinerVariable_Diffuse, true)));
	cds.overrideSpecularColor = ((pass.constantSlots[CombinerConstantSlot_Specular] != CombinerValue()) && (pass.constantSlots[CombinerConstantSlot_Specular] != CombinerValue(CombinerVariable_Specular)));
	cds.overrideSpecularAlpha = ((pass.constantSlots[CombinerConstantSlot_SpecularAlpha] != CombinerValue()) && (pass.constantSlots[CombinerConstantSlot_SpecularAlpha] != CombinerValue(CombinerVariable_Specular, true)));
	
	if(cds.overrideDiffuseColor && cds.overrideDiffuseAlpha)
	{
		cds.diffuse = (D3DCOLOR)MergeColorAndAlpha(pass.constantSlots[CombinerConstantSlot_Diffuse].ComputeColor(primitive, environment), pass.constantSlots[CombinerConstantSlot_DiffuseAlpha].ComputeAlpha(primitive, environment));
	}
	else if(cds.overrideDiffuseColor)
	{
		cds.diffuse = (D3DCOLOR)pass.constantSlots[CombinerConstantSlot_Diffuse].ComputeColor(primitive, environment);
	}
	else if(cds.overrideDiffuseAlpha)
	{
		cds.diffuse = (D3DCOLOR)CombinerColor(pass.constantSlots[CombinerConstantSlot_DiffuseAlpha].ComputeAlpha(primitive, environment));
	}

	if(cds.overrideSpecularColor && cds.overrideSpecularAlpha)
	{
		cds.specular = (D3DCOLOR)MergeColorAndAlpha(pass.constantSlots[CombinerConstantSlot_Specular].ComputeColor(primitive, environment), pass.constantSlots[CombinerConstantSlot_SpecularAlpha].ComputeAlpha(primitive, environment));
	}
	else if(cds.overrideDiffuseColor)
	{
		cds.specular = (D3DCOLOR)pass.constantSlots[CombinerConstantSlot_Specular].ComputeColor(primitive, environment);
	}
	else if(cds.overrideDiffuseAlpha)
	{
		cds.specular = (D3DCOLOR)CombinerColor(pass.constantSlots[CombinerConstantSlot_SpecularAlpha].ComputeAlpha(primitive, environment));
	}

	return S_OK;
}

bool D3DRender::ApplyMuxCombineMode(CombinerInfo& info, int passNum, CombinerVertexModificationInfo& cds)
{
	if(passNum >= info.passes.size())
	{
		// TODO: is this necessary?
		m_pD3DDev->SetRenderState( D3DRENDERSTATE_SRCBLEND,   D3DBLEND_SRCALPHA);
		m_pD3DDev->SetRenderState( D3DRENDERSTATE_DESTBLEND,  D3DBLEND_INVSRCALPHA);
		return false;
	}

	CombinerPass& pass = info.passes[passNum];

	ApplyCombinerPass(pass, cds);

	return true;
}

struct Replace01InMuxData
{
	bool prim0;
	bool prim1;
	bool primAlpha0;
	bool primAlpha1;
	bool env0;
	bool env1;
	bool envAlpha0;
	bool envAlpha1;
};

UINT Replace01InMux32(UINT value, const Replace01InMuxData& data)
{
	switch(value & 0x1f)
	{
		case MuxRawValue_Primitive:
			if(data.prim0)
				return MuxRawValue_32_0;
			if(data.prim1)
				return MuxRawValue_1;
			break;
		case MuxRawValue_Environment:
			if(data.env0)
				return MuxRawValue_32_0;
			if(data.env1)
				return MuxRawValue_1;
			break;
		case MuxRawValue_PrimitiveAlpha:
			if(data.primAlpha0)
				return MuxRawValue_32_0;
			if(data.primAlpha1)
				return MuxRawValue_1;
			break;
		case MuxRawValue_EnvironmentAlpha:
			if(data.envAlpha0)
				return MuxRawValue_32_0;
			if(data.envAlpha1)
				return MuxRawValue_1;
			break;
	}
	return value & 0x1f;
}

UINT Replace01InMux16(UINT value, const Replace01InMuxData& data)
{
	switch(value & 0xf)
	{
		case MuxRawValue_Primitive:
			if(data.prim0)
				return MuxRawValue_16_0;
			if(data.prim1)
				return MuxRawValue_1;
			break;
		case MuxRawValue_Environment:
			if(data.env0)
				return MuxRawValue_16_0;
			if(data.env1)
				return MuxRawValue_1;
			break;
		case MuxRawValue_PrimitiveAlpha:
			if(data.primAlpha0)
				return MuxRawValue_16_0;
			if(data.primAlpha1)
				return MuxRawValue_1;
			break;
		case MuxRawValue_EnvironmentAlpha:
			if(data.envAlpha0)
				return MuxRawValue_16_0;
			if(data.envAlpha1)
				return MuxRawValue_1;
			break;
	}
	return value & 0xf;
}

UINT Replace01InMux8(UINT value, const Replace01InMuxData& data)
{
	switch(value & 0x7)
	{
		case MuxRawValue_Primitive:
			if(data.prim0)
				return MuxRawValue_8_0;
			if(data.prim1)
				return MuxRawValue_1;
			break;
		case MuxRawValue_Environment:
			if(data.env0)
				return MuxRawValue_8_0;
			if(data.env1)
				return MuxRawValue_1;
			break;
	}
	return value & 0x7;
}

UINT Replace01InMux8InAlpha(UINT value, const Replace01InMuxData& data)
{
	switch(value & 0x7)
	{
		case MuxRawValue_Primitive:
			if(data.primAlpha0)
				return MuxRawValue_8_0;
			if(data.primAlpha1)
				return MuxRawValue_1;
			break;
		case MuxRawValue_Environment:
			if(data.envAlpha0)
				return MuxRawValue_8_0;
			if(data.envAlpha1)
				return MuxRawValue_1;
			break;
	}
	return value & 0x7;
}

u64 Replace01InMux(u64 mux, const Replace01InMuxData& data)
{
	DWORD mux0 = (DWORD)(mux>>32);
	DWORD mux1 = (DWORD)(mux);
	
	mux0 =
		(Replace01InMux16(mux0 >> 20, data)			<< 20) | // c0 a
		(Replace01InMux32(mux0 >> 15, data)			<< 15) | // c0 c
		(Replace01InMux8InAlpha(mux0 >> 12, data)	<< 12) | // a0 a
		(Replace01InMux8InAlpha(mux0 >> 9, data)	<< 9) |  // a0 c
		(Replace01InMux16(mux0 >> 5, data)			<< 5) |  // c1 a
		(Replace01InMux32(mux0, data));				         // c1 c
	
	mux1 =
		(Replace01InMux16(mux1 >> 28, data)			<< 28) | // c0 b
		(Replace01InMux16(mux1 >> 24, data)			<< 24) | // c1 b
		(Replace01InMux8InAlpha(mux1 >> 21, data)	<< 21) | // a1 a
		(Replace01InMux8InAlpha(mux1 >> 18, data)	<< 18) | // a1 c
		(Replace01InMux8(mux1 >> 15, data)			<< 15) | // c0 d
		(Replace01InMux8InAlpha(mux1 >> 12, data)	<< 12) | // a0 b
		(Replace01InMux8InAlpha(mux1 >> 9, data)	<< 9) |  // a0 d
		(Replace01InMux8(mux1 >> 6, data)			<< 6) |  // c1 d
		(Replace01InMux8InAlpha(mux1 >> 3, data)	<< 3) |  // a1 b
		(Replace01InMux8InAlpha(mux1, data));                // a1 d

	return ((u64)mux0 << 32) | (u64)mux1;
}

#ifndef DAEDALUS_RELEASE_BUILD
map<u64, u64> muxSubstMap;
#endif

bool D3DRender::ApplyCombineMode(int passNum, CombinerVertexModificationInfo& cds)
{
	DWORD cycleType = m_dwCycleType << G_MDSFT_CYCLETYPE;
	if((cycleType == G_CYC_COPY) || (cycleType == G_CYC_FILL))
	{
		if(passNum > 0)
			return false;

		cds.overrideDiffuseColor = false;
		cds.overrideDiffuseAlpha = false;
		cds.overrideSpecularColor = false;
		cds.overrideSpecularAlpha = false;

		SetColorOp( 1, D3DTOP_DISABLE );
		SetAlphaOp( 1, D3DTOP_DISABLE );

		//m_pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
		//m_pD3DDev->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

		m_pD3DDev->SetRenderState( D3DRENDERSTATE_ALPHABLENDENABLE,   TRUE);
		m_pD3DDev->SetRenderState( D3DRENDERSTATE_ALPHATESTENABLE,   TRUE);
			
		//SetNumStages(0);

		if (cycleType == G_CYC_COPY)
		{		
			DL_PF("CycleType: Copy");

			cds.texturesCount = 1;
			cds.textureTypes[0] = TextureType_Texture0;

			m_pD3DDev->SetTexture( 0, m_lpsTexture );

			SetColorOp( 0, D3DTOP_SELECTARG1 );
			SetColorArg1( 0, D3DTA_TEXTURE );

			SetAlphaOp( 0, D3DTOP_SELECTARG1 );
			SetAlphaArg1( 0, D3DTA_TEXTURE );
		}	
		else
		{
			DL_PF("CycleType: Fill");

			cds.texturesCount = 0;

			m_pD3DDev->SetTexture( 0, NULL );

			// TODO: why not selectarg1?
			SetColorOp( 0, D3DTOP_SELECTARG2 );
			SetColorArg2( 0, D3DTA_DIFFUSE );

			SetAlphaOp( 0, D3DTOP_SELECTARG2 );
			SetAlphaArg2( 0, D3DTA_DIFFUSE );
		}
		return true;
	}
	else
	{
		if ((m_dwCycleType << G_MDSFT_CYCLETYPE) == G_CYC_1CYCLE)
		{
			DL_PF("CycleType: 1");
		}
		else
		{
			DL_PF("CycleType: 2");
		}

		// TODO: this should be done in the functions for handling combine mode, env and prim setting
		Replace01InMuxData replaceData;
		u32 primColor = (D3DCOLOR)m_dwPrimitiveColor & 0x00ffffff;
		replaceData.prim0 = (primColor == 0);
		replaceData.prim1 = (primColor == 0x00ffffff);

		u32 envColor = (D3DCOLOR)m_dwEnvColor & 0x00ffffff;
		replaceData.env0 = (envColor == 0);
		replaceData.env1 = (envColor == 0x00ffffff);

		u32 primAlpha = (D3DCOLOR)m_dwPrimitiveColor & 0xff000000;
		replaceData.primAlpha0 = (primAlpha == 0);
		replaceData.primAlpha1 = (primAlpha == 0xff000000);

		u32 envAlpha = (D3DCOLOR)m_dwEnvColor & 0xff000000;
		replaceData.envAlpha0 = (envAlpha == 0);
		replaceData.envAlpha1 = (envAlpha == 0xff000000);

		u64 mux = m_Mux;
		mux = Replace01InMux(mux, replaceData);

		#ifndef DAEDALUS_RELEASE_BUILD
		map<u64, u64>::const_iterator iter = muxSubstMap.find(mux);
		if(iter != muxSubstMap.end())
		{
			if(iter->second == 0)
			{
				return false;
			}
			else
			{
				mux = iter->second;
			}
		}
		#endif

		CombinerInfo& info = FindOrComputeCombinerInfoFromMux(mux);
		return ApplyMuxCombineMode(info, passNum, cds);
	}
}


